/* Process model C form file: DVMRP.pr.c */
/* Portions of this file copyright 1992, 1996 by MIL 3, Inc. */



/* OPNET system definitions */
#include <opnet.h>
#include "/disk2/gduan/qosip/gwy_dvmrp/DVMRP.pr.h"
FSM_EXT_DECS



/* Header block */
/**
***
*** Header file of DVMRP
*** Author: Gang Duan, Jiemei Ma
*** Date:   March, 1998
***
**/
 
#define PROBE_INTR      1
#define REPORT_INTR     2
#define GRAFT_INTR      3
#define PRUNE_INTR      4
#define CALC_INTR	5
#define TIMER		6

#define PROBE_PKCODE	1
#define REPORT_PKCODE 	2
#define PRUNE_PKCODE	7
#define GRAFT_PKCODE	8
#define GRAFTACK_PKCODE 9
 
#define ARR     ( op_intrpt_type() == OPC_INTRPT_STRM )
#define PROBE   ( op_intrpt_type() == OPC_INTRPT_SELF && op_intrpt_code() == PROBE_INTR )
#define REPORT  ( op_intrpt_type() == OPC_INTRPT_SELF && op_intrpt_code() == REPORT_INTR )
#define PRUNE   ( op_intrpt_type() == OPC_INTRPT_REMOTE && op_intrpt_code() == PRUNE_INTR )
#define GRAFT   ( op_intrpt_type() == OPC_INTRPT_REMOTE && op_intrpt_code() == GRAFT_INTR )
#define CALC    ( op_intrpt_type() == OPC_INTRPT_REMOTE && op_intrpt_code() == CALC_INTR )
#define PULSE	( op_intrpt_type() == OPC_INTRPT_SELF && op_intrpt_code() == TIMER )


/**
*** Protocol ID's
**/
#define UDP_PROTOCOL_ID         20
#define TCP_PROTOCOL_ID         12
#define IGMP_PROTOCOL_ID        9
#define DVMRP_PROTOCOL_ID	13

#define IPSTREAM                0
#define IGMPSTREAM              1
#define DEF_STREAM              110

#define AllDVMRPRoutersNetNum     400
#define AllDVMRPRoutersNodeNum    5
#define MULTICAST_NET           500

#define REPORT_INTERVAL		60
#define PROBE_INTERVAL		10	


#define TRUE                    1
#define FALSE			0

#define ON			1 
#define OFF			0 
#define VL                      if(ON)
#define PROBE_DEBUG		if(OFF)
#define INIT_DEBUG		if(OFF)
#define REPORT_DEBUG		if(OFF)
#define PRUNE_DEBUG		if(OFF)
#define GRAFT_DEBUG		if(OFF)
#define CAL_DEBUG		if(OFF)


/**
*** Maximum Number Limitations on the data structures numbers
**/
#define MAX_NO_NEIGHBORS                10
#define MAX_NO_AREAS                    5
#define MAX_NO_GRPS             30
#define MAX_NO_LINKS                    10

#define MAX_NO_ROUTES	100
#define MAX_NO_FORWARDS 200


#define GraftResendTimer	5
#define PruneLifetime		20

/**
*** IP Address Data Structure
**/
typedef struct{
        int NetNum;
        int NodeNum;
}IPAddress;

typedef enum{ UP, DOWN }RouteStatus;

typedef struct{
        IPAddress IntfIPAddr;
        IPAddress NeighborRtrId;
	RouteStatus flag;
	int tick;
}Linkinfo;



typedef struct{
	int NoOfNeighbors;
	IPAddress neighbor[MAX_NO_LINKS];
}ProbeStruct;


typedef struct{
	IPAddress source;
	IPAddress upstream;
	int metric;
	RouteStatus flag;
	int downstream[MAX_NO_LINKS];
	int tick;
}RouteInfo;

typedef struct{
	int NoOfRoutes;	
	IPAddress source[MAX_NO_ROUTES];
	IPAddress upstream[MAX_NO_ROUTES];
	int metric[MAX_NO_ROUTES];
}ReportStruct;


typedef struct{
        IPAddress source;
	int group;
        IPAddress upstream;
        int metric;
        RouteStatus flag;
        int downstream[MAX_NO_LINKS];
	RouteStatus downflag[MAX_NO_LINKS];
        int downtick[MAX_NO_LINKS];
	int grafttick;
}ForwardInfo;


typedef struct{
	IPAddress source;
	int group;
	int lifetime;
}PruneStruct; 


typedef struct{
	IPAddress source;
	int group;
}GraftStruct;


typedef struct{
	int interface[MAX_NO_LINKS];
	int mtu[MAX_NO_LINKS];
	int infValid;
}mc_id;



void initialize();
void probePkPro( ProbeStruct * );
void reportPkPro( ReportStruct * );
void prunePkPro( PruneStruct * );
void graftPkPro( GraftStruct * );
void graftAckPkPro( GraftStruct * );

int addDownStrmFwding( IPAddress source, int downStrm );
int delDownStrmFwding( IPAddress source, int downStrm );

void prune( IPAddress dest, IPAddress source, int group, int lifetime );
void graft( IPAddress dest, IPAddress source, int group );
void graftAck( int dest, IPAddress source, int group );
int ip2stream(IPAddress up);
void print_ForwardingTable();
#if !defined (VOSD_NO_FIN)
#undef	BIN
#undef	BOUT
#define	BIN		_fstack_local_info.last_line_passed = __LINE__ - _block_origin;
#define	BOUT	BIN
#define	BINIT	_fstack_local_info.last_line_passed = 0; _block_origin = __LINE__;
#else
#define	BINIT
#endif /* #if !defined (VOSD_NO_FIN) */



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	Objid                   		sv_my_id;
	Objid                   		sv_ip_objid;
	Objid                   		sv_igmp_objid;
	Objid                   		sv_sub_netid;
	Distribution*           		sv_report_dist;
	int                     		sv_AreaId;
	int                     		sv_NoOfLinks;
	int                     		sv_Indep_Rou;
	int                     		sv_IS_ROUTER;
	int                     		sv_IS_DESIGNATED_ROUTER;
	int                     		sv_IS_AREA_BORDER_ROUTER;
	int                     		sv_net_num0;
	int                     		sv_node_num0;
	int                     		sv_net_num1;
	int                     		sv_node_num1;
	int                     		sv_net_num2;
	int                     		sv_node_num2;
	int                     		sv_net_num3;
	int                     		sv_node_num3;
	int                     		sv_net_num4;
	int                     		sv_node_num4;
	int                     		sv_net_num5;
	int                     		sv_node_num5;
	int                     		sv_net_num6;
	int                     		sv_node_num6;
	int                     		sv_net_num7;
	int                     		sv_node_num7;
	int                     		sv_net_num8;
	int                     		sv_node_num8;
	int                     		sv_net_num9;
	int                     		sv_node_num9;
	int                     		sv_in_strm;
	int                     		sv_ip_strm;
	int                     		sv_src_net;
	int                     		sv_src_node;
	int                     		sv_grp_id;
	int                     		sv_dgram_id;
	int                     		sv_NoOfNeighborElements;
	int                     		sv_InstNoLsa;
	Linkinfo                		sv_LinksInfo[MAX_NO_NEIGHBORS];
	RouteInfo               		sv_RoutingTable[MAX_NO_ROUTES];
	ForwardInfo             		sv_ForwardingTable[MAX_NO_FORWARDS];
	} DVMRP_state;

#define pr_state_ptr            		((DVMRP_state*) SimI_Mod_State_Ptr)
#define my_id                   		pr_state_ptr->sv_my_id
#define ip_objid                		pr_state_ptr->sv_ip_objid
#define igmp_objid              		pr_state_ptr->sv_igmp_objid
#define sub_netid               		pr_state_ptr->sv_sub_netid
#define report_dist             		pr_state_ptr->sv_report_dist
#define AreaId                  		pr_state_ptr->sv_AreaId
#define NoOfLinks               		pr_state_ptr->sv_NoOfLinks
#define Indep_Rou               		pr_state_ptr->sv_Indep_Rou
#define IS_ROUTER               		pr_state_ptr->sv_IS_ROUTER
#define IS_DESIGNATED_ROUTER    		pr_state_ptr->sv_IS_DESIGNATED_ROUTER
#define IS_AREA_BORDER_ROUTER   		pr_state_ptr->sv_IS_AREA_BORDER_ROUTER
#define net_num0                		pr_state_ptr->sv_net_num0
#define node_num0               		pr_state_ptr->sv_node_num0
#define net_num1                		pr_state_ptr->sv_net_num1
#define node_num1               		pr_state_ptr->sv_node_num1
#define net_num2                		pr_state_ptr->sv_net_num2
#define node_num2               		pr_state_ptr->sv_node_num2
#define net_num3                		pr_state_ptr->sv_net_num3
#define node_num3               		pr_state_ptr->sv_node_num3
#define net_num4                		pr_state_ptr->sv_net_num4
#define node_num4               		pr_state_ptr->sv_node_num4
#define net_num5                		pr_state_ptr->sv_net_num5
#define node_num5               		pr_state_ptr->sv_node_num5
#define net_num6                		pr_state_ptr->sv_net_num6
#define node_num6               		pr_state_ptr->sv_node_num6
#define net_num7                		pr_state_ptr->sv_net_num7
#define node_num7               		pr_state_ptr->sv_node_num7
#define net_num8                		pr_state_ptr->sv_net_num8
#define node_num8               		pr_state_ptr->sv_node_num8
#define net_num9                		pr_state_ptr->sv_net_num9
#define node_num9               		pr_state_ptr->sv_node_num9
#define in_strm                 		pr_state_ptr->sv_in_strm
#define ip_strm                 		pr_state_ptr->sv_ip_strm
#define src_net                 		pr_state_ptr->sv_src_net
#define src_node                		pr_state_ptr->sv_src_node
#define grp_id                  		pr_state_ptr->sv_grp_id
#define dgram_id                		pr_state_ptr->sv_dgram_id
#define NoOfNeighborElements    		pr_state_ptr->sv_NoOfNeighborElements
#define InstNoLsa               		pr_state_ptr->sv_InstNoLsa
#define LinksInfo               		pr_state_ptr->sv_LinksInfo
#define RoutingTable            		pr_state_ptr->sv_RoutingTable
#define ForwardingTable         		pr_state_ptr->sv_ForwardingTable

/* This macro definition will define a local variable called	*/
/* "op_sv_ptr" in each function containing a FIN statement.	*/
/* This variable points to the state variable data structure,	*/
/* and can be used from a C debugger to display their values.	*/
#undef FIN_PREAMBLE
#define FIN_PREAMBLE	DVMRP_state *op_sv_ptr = pr_state_ptr;




/* Process model interrupt handling procedure */



void
DVMRP ()
	{
	int data_len;
	
	int protocolid;
	int pkCode;
	
	Packet *ReceivedDVMRPPkptr;
	Packet *ReceivedPkptr;
	
	ReportStruct *rcvReportPtr;
	ProbeStruct *rcvProbePtr;
	PruneStruct *rcvPrunePtr;
	GraftStruct *rcvGraftPtr;
	
	int _block_origin = 0;


	FSM_ENTER (DVMRP)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (Init) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "Init", "DVMRP () [Init enter execs]")
			{
			int i;
			
			my_id = op_id_self();
			
			IS_ROUTER = 0;
			IS_DESIGNATED_ROUTER = 0;
			IS_AREA_BORDER_ROUTER = 0;
			
			op_ima_obj_attr_get( my_id, "AreaId", &AreaId );
			op_ima_obj_attr_get( my_id, "IsRouter", &IS_ROUTER );
			op_ima_obj_attr_get( my_id, "AreaBorderRouter", &IS_AREA_BORDER_ROUTER );
			op_ima_obj_attr_get( my_id, "DesignatedRouter", &IS_DESIGNATED_ROUTER );
			
			ip_objid = op_topo_assoc(my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, IPSTREAM);
			igmp_objid = op_topo_assoc(my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, IGMPSTREAM);
			
			/**
			***  Using this IP layer Object Id get all the required attributes from the IP layer
			**/
			op_ima_obj_attr_get( ip_objid, "net_num0", &net_num0 );
			op_ima_obj_attr_get( ip_objid, "node_num0", &node_num0 );
			op_ima_obj_attr_get( ip_objid, "net_num1", &net_num1 );
			op_ima_obj_attr_get( ip_objid, "node_num1", &node_num1 );
			op_ima_obj_attr_get( ip_objid, "net_num2", &net_num2 );
			op_ima_obj_attr_get( ip_objid, "node_num2", &node_num2 );
			op_ima_obj_attr_get( ip_objid, "net_num3", &net_num3 );
			op_ima_obj_attr_get( ip_objid, "node_num3", &node_num3 );
			op_ima_obj_attr_get( ip_objid, "net_num4", &net_num4 );
			op_ima_obj_attr_get( ip_objid, "node_num4", &node_num4 );
			op_ima_obj_attr_get( ip_objid, "net_num5", &net_num5 );
			op_ima_obj_attr_get( ip_objid, "node_num5", &node_num5 );
			op_ima_obj_attr_get( ip_objid, "net_num6", &net_num6 );
			op_ima_obj_attr_get( ip_objid, "node_num6", &node_num6 );
			op_ima_obj_attr_get( ip_objid, "net_num7", &net_num7 );
			op_ima_obj_attr_get( ip_objid, "node_num7", &node_num7 );
			op_ima_obj_attr_get( ip_objid, "net_num8", &net_num8 );
			op_ima_obj_attr_get( ip_objid, "node_num8", &node_num8 );
			op_ima_obj_attr_get( ip_objid, "net_num9", &net_num9 );
			op_ima_obj_attr_get( ip_objid, "node_num9", &node_num9 );
			op_ima_obj_attr_get( ip_objid, "Indep_Rou", &Indep_Rou );
			
			
			if( !Indep_Rou )
			        sub_netid=op25_topo_parent( op25_topo_parent(my_id) );
			else
			{
			        sub_netid=op25_topo_parent( my_id );
			}
			
			
			src_node = node_num0;
			src_net = net_num0;
			
			NoOfLinks = op_topo_assoc_count( sub_netid,OPC_TOPO_ASSOC_IN,OPC_OBJTYPE_LKSIMP );
			
			initialize();
			
			VL INIT_DEBUG  
			{
			        printf("--------------  INIT BLOCK PROCESSING  ----------------\n");
			        printf("Router(%d.%d) Link Information \n", net_num0, node_num0 );
			        printf("Number of Links is %d\n ", NoOfLinks);
			        for( i = 0; i < NoOfLinks; i++ )
			        {
			/*                printf("Network_Router[%d]=%d.%d  ",i,
						Network_Routers[i].NetNum,Network_Routers[i].NodeNum);
			*/
			                printf("LinksInfo[%d].NeighborRtrId=%d.%d  ",i,
			                        LinksInfo[i].NeighborRtrId.NetNum, LinksInfo[i].NeighborRtrId.NodeNum );
			                printf("LinksInfo[%d].IntfIPAddr=%d.%d \n ",i,
			                        LinksInfo[i].IntfIPAddr.NetNum, LinksInfo[i].IntfIPAddr.NodeNum );
			        }
			        printf("------------  END INIT BLOCK PROCESSING  --------------\n\n");
			}
			
			report_dist = op_dist_load( "uniform_int", REPORT_INTERVAL * .5, REPORT_INTERVAL * 1.5 );
			
			op_intrpt_schedule_self( op_sim_time() + 1, PROBE_INTR );
			op_intrpt_schedule_self( op_sim_time() + 2, REPORT_INTR );
			op_intrpt_schedule_self( op_sim_time() + 1, TIMER );
			
			}


		/** state (Init) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "Init", "DVMRP () [Init exit execs]")
			{
			
			}


		/** state (Init) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (Idle) enter executives **/
		FSM_STATE_ENTER_UNFORCED (1, state1_enter_exec, "Idle", "DVMRP () [Idle enter execs]")
			{
			
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (3,DVMRP)


		/** state (Idle) exit executives **/
		FSM_STATE_EXIT_UNFORCED (1, state1_exit_exec, "Idle", "DVMRP () [Idle exit execs]")
			{
			}


		/** state (Idle) transition processing **/
		FSM_INIT_COND (PROBE)
		FSM_TEST_COND (REPORT)
		FSM_TEST_COND (ARR)
		FSM_TEST_COND (GRAFT)
		FSM_TEST_COND (CALC)
		FSM_TEST_COND (PULSE)
		FSM_TEST_COND (PRUNE)
		FSM_TEST_LOGIC ("Idle")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 2, state2_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 5, state5_enter_exec, ;)
			FSM_CASE_TRANSIT (2, 4, state4_enter_exec, ;)
			FSM_CASE_TRANSIT (3, 3, state3_enter_exec, ;)
			FSM_CASE_TRANSIT (4, 6, state6_enter_exec, ;)
			FSM_CASE_TRANSIT (5, 7, state7_enter_exec, ;)
			FSM_CASE_TRANSIT (6, 8, state8_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (Probe_Send) enter executives **/
		FSM_STATE_ENTER_FORCED (2, state2_enter_exec, "Probe_Send", "DVMRP () [Probe_Send enter execs]")
			{
			/** 
			*** Send Probe Message every PROBE_INTERVAL time
			*** Probe Message includes a list of the neighbors' IP addresses
			**/
			
			Packet *probe_pkptr, *ip_pkptr;
			int data_len;
			ProbeStruct *ps;
			int i;
			
			VL PROBE_DEBUG { 
			printf("   (%d.%d) DVMRP probe\n", net_num0, node_num0);
			printf("--------------------------\n");
			}
			
			probe_pkptr = op_pk_create_fmt( "dvmrp_pkfmt" );
			op25_pk_nfd_set( probe_pkptr, "Code", PROBE_PKCODE );
			
			ps = (ProbeStruct *) op_prg_mem_alloc(sizeof(ProbeStruct));
			for( i = 0; i < NoOfLinks; i++ )
			{
				ps->neighbor[i].NetNum = LinksInfo[i].NeighborRtrId.NetNum;
				ps->neighbor[i].NodeNum = LinksInfo[i].NeighborRtrId.NodeNum;
			VL PROBE_DEBUG printf(" %d.%d\n",ps->neighbor[i].NetNum, ps->neighbor[i].NodeNum);
			}
			VL  PROBE_DEBUG printf("--------------------------\n");
			ps->NoOfNeighbors = NoOfLinks;
			
			op25_pk_nfd_set( probe_pkptr ,"Data", ps, op_prg_mem_alloc,
			                 op_prg_mem_free, sizeof(ProbeStruct) );
			
			data_len = op_pk_total_size_get( probe_pkptr ) / 8;
			
			ip_pkptr = op_pk_create_fmt( "ip_dgram" );
			op25_pk_nfd_set( ip_pkptr, "data", probe_pkptr );
			op_pk_bulk_size_set( ip_pkptr, data_len * 8 );
			
			op25_pk_nfd_set( ip_pkptr, "prot_id", DVMRP_PROTOCOL_ID );
			op25_pk_nfd_set( ip_pkptr, "dest_net", AllDVMRPRoutersNetNum );
			op25_pk_nfd_set( ip_pkptr, "dest_node", AllDVMRPRoutersNodeNum );
			op25_pk_nfd_set( ip_pkptr, "ttl", 2 );
			
			op25_pk_nfd_set( ip_pkptr, "src_net", net_num0 );
			op25_pk_nfd_set( ip_pkptr, "src_node", node_num0 );
			
			op25_pk_nfd_set( ip_pkptr, "orig_len", data_len );
			op25_pk_nfd_set( ip_pkptr, "frag_len", data_len );
			op25_pk_nfd_set( ip_pkptr, "frag", 0 );
			op25_pk_nfd_set( ip_pkptr, "ident", dgram_id++ );
			
			op_pk_send( ip_pkptr, IPSTREAM );
			
			if (op_sim_time() < 10.00 ) 
			    op_intrpt_schedule_self( op_sim_time() + 2, PROBE_PKCODE );
			else 
			    op_intrpt_schedule_self( op_sim_time() + PROBE_INTERVAL, PROBE_PKCODE );
			
			 
			
			}


		/** state (Probe_Send) exit executives **/
		FSM_STATE_EXIT_FORCED (2, state2_exit_exec, "Probe_Send", "DVMRP () [Probe_Send exit execs]")
			{
			
			}


		/** state (Probe_Send) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (Graft_Send) enter executives **/
		FSM_STATE_ENTER_FORCED (3, state3_enter_exec, "Graft_Send", "DVMRP () [Graft_Send enter execs]")
			{
			int group;
			int i;
			
			printf(" (%d.%d) Graft state\n", net_num0, node_num0);
			/* Only GWY will reach here */
			
			group = *((int*) op_ima_obj_svar_get(igmp_objid, "grp_id"));
			
			printf(" group number : %d\n", group);
			
			/* check forwarding table, send graft to each (source, group) pair */
			i = 0;
			while( ForwardingTable[i].source.NetNum != 0 &&
			       ForwardingTable[i].source.NodeNum != 0 )
			{
				if ( ForwardingTable[i].group == group &&
			             ( ForwardingTable[i].source.NetNum != net_num0 ||
			               ForwardingTable[i].source.NodeNum != node_num0 ) )
				{
					graft( ForwardingTable[i].upstream, 
					       ForwardingTable[i].source, group );
					ForwardingTable[i].grafttick = GraftResendTimer;
				}
				i++;
			}
			
			
			}


		/** state (Graft_Send) exit executives **/
		FSM_STATE_EXIT_FORCED (3, state3_exit_exec, "Graft_Send", "DVMRP () [Graft_Send exit execs]")
			{
			}


		/** state (Graft_Send) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (Arr_Pkt) enter executives **/
		FSM_STATE_ENTER_FORCED (4, state4_enter_exec, "Arr_Pkt", "DVMRP () [Arr_Pkt enter execs]")
			{
			in_strm = op_intrpt_strm();
			
			ReceivedPkptr = op_pk_get( in_strm );
			
			op_pk_nfd_get( ReceivedPkptr, "strm", &ip_strm );
			op_pk_nfd_get( ReceivedPkptr, "prot_id", &protocolid );
			op_pk_nfd_get( ReceivedPkptr, "src_net", &src_net );
			op_pk_nfd_get( ReceivedPkptr, "src_node", &src_node );
			
			if ( protocolid == DVMRP_PROTOCOL_ID )
			{
				op_pk_nfd_get( ReceivedPkptr, "data", &ReceivedDVMRPPkptr );
				op_pk_nfd_get( ReceivedDVMRPPkptr, "Code", &pkCode );
				switch( pkCode )
				{
					case PROBE_PKCODE:
						op_pk_nfd_get( ReceivedDVMRPPkptr, "Data", &rcvProbePtr );
						probePkPro( rcvProbePtr );
						break;
					case REPORT_PKCODE:
						op_pk_nfd_get( ReceivedDVMRPPkptr, "Data", &rcvReportPtr );
						reportPkPro( rcvReportPtr );
						break;
					case PRUNE_PKCODE:
						op_pk_nfd_get( ReceivedDVMRPPkptr, "Data", &rcvPrunePtr );
						prunePkPro( rcvPrunePtr );
						break;
			                case GRAFT_PKCODE:
			                        op_pk_nfd_get( ReceivedDVMRPPkptr, "Data", &rcvGraftPtr );
			                        graftPkPro( rcvGraftPtr );
			                        break;
			                case GRAFTACK_PKCODE:
			                        op_pk_nfd_get( ReceivedDVMRPPkptr, "Data", &rcvGraftPtr );
			                        graftAckPkPro( rcvGraftPtr );
			                        break;
				}
			} 
			
			op_pk_destroy( ReceivedPkptr );
			
			}


		/** state (Arr_Pkt) exit executives **/
		FSM_STATE_EXIT_FORCED (4, state4_exit_exec, "Arr_Pkt", "DVMRP () [Arr_Pkt exit execs]")
			{
			}


		/** state (Arr_Pkt) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (Report_Send) enter executives **/
		FSM_STATE_ENTER_FORCED (5, state5_enter_exec, "Report_Send", "DVMRP () [Report_Send enter execs]")
			{
			/**
			*** Send Report Message
			*** Not like the packet format in DVMRP3, instead of adding 32 to the metric back to
			*** the upstream, we add another field specifing the upstream IP.
			*** The shortcoming is the bigger size of the report message.
			*** We couldn't figure out how to implement the method in DVMRP3.
			**/
			
			Packet *report_pkptr, *ip_pkptr;
			int data_len;
			ReportStruct *rs;
			int i=0, report_interval;
			
			
			report_pkptr = op_pk_create_fmt( "dvmrp_pkfmt" );
			op25_pk_nfd_set( report_pkptr, "Code", REPORT_PKCODE );
			
			rs = (ReportStruct *) op_prg_mem_alloc(sizeof(ReportStruct));
			memset( rs, 0, sizeof(ReportStruct));
			
			while( RoutingTable[i].source.NetNum != 0 &&
			       RoutingTable[i].source.NodeNum != 0 && i<MAX_NO_ROUTES)
			{
			        rs->source[i].NetNum = RoutingTable[i].source.NetNum;
			        rs->source[i].NodeNum = RoutingTable[i].source.NodeNum;
			        rs->upstream[i].NetNum = RoutingTable[i].upstream.NetNum;
			        rs->upstream[i].NodeNum = RoutingTable[i].upstream.NodeNum;
				rs->metric[i] = RoutingTable[i].metric;
				i++;
				rs->NoOfRoutes++;
			}
			
			VL REPORT_DEBUG {
			    printf("            (%d.%d) DVMRP report\n", net_num0, node_num0);
			    printf("------------------------------------------\n");
			    printf("NoOfRouters = %d\n", rs->NoOfRoutes);
			    for (i=0; i< rs->NoOfRoutes; i++) {
				printf("source (%d.%d), upstream (%d.%d) metric %d\n",  
					rs->source[i].NetNum, rs->source[i].NodeNum,
					rs->upstream[i].NetNum, rs->upstream[i].NodeNum,
					rs->metric[i]);
			    }
			    printf("------------------------------------------\n");
			}
			
			op25_pk_nfd_set( report_pkptr ,"Data", rs, op_prg_mem_alloc,
			                 op_prg_mem_free, sizeof(ReportStruct) );
			
			data_len = op_pk_total_size_get( report_pkptr ) / 8;
			
			ip_pkptr = op_pk_create_fmt( "ip_dgram" );
			op25_pk_nfd_set( ip_pkptr, "data", report_pkptr );
			op_pk_bulk_size_set( ip_pkptr, data_len * 8 );
			
			op25_pk_nfd_set( ip_pkptr, "dest_net", AllDVMRPRoutersNetNum );
			op25_pk_nfd_set( ip_pkptr, "dest_node", AllDVMRPRoutersNodeNum );
			op25_pk_nfd_set( ip_pkptr, "ttl", 2 );
			
			op25_pk_nfd_set( ip_pkptr, "src_net", net_num0 );
			op25_pk_nfd_set( ip_pkptr, "src_node", node_num0 );
			
			op25_pk_nfd_set( ip_pkptr, "orig_len", data_len );
			op25_pk_nfd_set( ip_pkptr, "frag_len", data_len );
			op25_pk_nfd_set( ip_pkptr, "frag", 0 );
			op25_pk_nfd_set( ip_pkptr, "ident", dgram_id++ );
			op25_pk_nfd_set( ip_pkptr, "prot_id", DVMRP_PROTOCOL_ID );
			
			op_pk_send( ip_pkptr, IPSTREAM );
			
			if (op_sim_time() < 10.00 ) {
			    op_intrpt_schedule_self( op_sim_time() + 1, REPORT_PKCODE );
			} else {
			    report_interval = op_dist_outcome( report_dist );
			    op_intrpt_schedule_self( op_sim_time() + report_interval, REPORT_PKCODE );
			}
			
			}


		/** state (Report_Send) exit executives **/
		FSM_STATE_EXIT_FORCED (5, state5_exit_exec, "Report_Send", "DVMRP () [Report_Send exit execs]")
			{
			
			}


		/** state (Report_Send) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (Route_Calc) enter executives **/
		FSM_STATE_ENTER_FORCED (6, state6_enter_exec, "Route_Calc", "DVMRP () [Route_Calc enter execs]")
			{
			int group;
			int s_net;
			int s_node;
			int fwd, FoundFwd, FoundGroup;
			int rt, i;
			ForwardInfo *finfo;
			RouteInfo *rinfo;
			int strm;
			mc_id *mc_tab;
			
			group = *((int*)op_ima_obj_svar_get( ip_objid, "GrpId" ));
			s_net = *((int*) op_ima_obj_svar_get(ip_objid, "SourceNet"));
			s_node = *((int*) op_ima_obj_svar_get(ip_objid, "SourceNode"));
			
			VL CAL_DEBUG {
			    printf("--- (%d.%d) create item in Forwarding Table for Source (%d.%d) Grp (%d) ---\n",
				   net_num0, node_num0, s_net, s_node, group);
			    printf("Before :\n");
			    print_ForwardingTable();
			    printf("\n\n");
			}
			
			/* Check if (source, group) exists in Forwarding table */
			fwd = 0;
			FoundFwd = 0;
			while( ForwardingTable[fwd].source.NetNum != 0 &&
			       ForwardingTable[fwd].source.NodeNum != 0 )
			{
				if ( ForwardingTable[fwd].source.NetNum == s_net &&
				     ForwardingTable[fwd].group == group )
				{
					FoundFwd = 1;
					break;
				}
				fwd++;
			}
			
			/* if not exist, copy an entry from Routingtable */
			if ( FoundFwd == 0 && fwd < MAX_NO_FORWARDS )
			{
				rt = 0;
				while( RoutingTable[rt].source.NetNum != 0 &&
				       RoutingTable[rt].source.NodeNum != 0 )
				{
					if ( RoutingTable[rt].source.NetNum == s_net) 
					{
						finfo = &ForwardingTable[fwd];
						rinfo = &RoutingTable[rt];
			
						finfo -> source.NetNum = rinfo -> source.NetNum;
						finfo -> source.NodeNum = rinfo -> source.NodeNum;
						finfo -> group = group;
						finfo -> upstream.NetNum = rinfo -> upstream.NetNum;
			                        finfo -> upstream.NodeNum = rinfo -> upstream.NodeNum;
						finfo -> metric = rinfo -> metric;
						finfo -> flag = rinfo -> flag;
						for( i = 0; i < NoOfLinks; i++ )
						{
							finfo -> downstream[i] = rinfo -> downstream[i];
							finfo -> downflag[i] = UP;
							finfo -> downtick[i] = -1;
						}
						finfo -> grafttick = -1;
			
						break;
					}
					rt++;
				}
			}
			 
			VL CAL_DEBUG {
			    printf("After :\n");
			    print_ForwardingTable();
			    printf("-----------------------------------------------------------------\n");
			}
			
			/* construct MC_Table using fouwarding table */
			mc_tab = (mc_id*)op_ima_obj_svar_get( ip_objid, "mc_table" );
			
			for( i = 0; i < NoOfLinks; i++ )
			{
				strm = ForwardingTable[fwd].downstream[i];
				if ( ForwardingTable[fwd].downflag[i] == UP )
					mc_tab[group].interface[strm] = 1;
				else
					mc_tab[group].interface[strm] = 0;
			}
			
			mc_tab[group].infValid = s_net;
			
			if (IS_ROUTER == 0) {
			    for( i = 0; i < NoOfLinks; i++ )
			    {
			        FoundGroup = 0;
			VL CAL_DEBUG {
			    printf(" mc_table: (grp %d, interface %d) : %d\n", group, i, mc_tab[group].interface[i]);
			}
			        if( mc_tab[group].interface[i] == 1)
			        {
			                FoundGroup = 1;
			                break;
			        }
			    }
			
			    if ( FoundGroup == 0 )
			    {
				i=0;
			        while( ForwardingTable[i].source.NetNum != 0 &&
			                ForwardingTable[i].source.NodeNum != 0 )
			        {
			                if ( ForwardingTable[i].group == group &&
			                     ( ForwardingTable[i].source.NetNum != net_num0 ||
			                       ForwardingTable[i].source.NodeNum != node_num0 ) )
			                {
			                        prune( ForwardingTable[i].upstream,
			                                ForwardingTable[i].source, group, PruneLifetime );
			                        ForwardingTable[i].downflag[0] == DOWN;
			                        ForwardingTable[i].downtick[0] = 0;
			                        ForwardingTable[i].grafttick = -1;
			                }
			                i++;
			        }
			    }
			
			}
			
			
			}


		/** state (Route_Calc) exit executives **/
		FSM_STATE_EXIT_FORCED (6, state6_exit_exec, "Route_Calc", "DVMRP () [Route_Calc exit execs]")
			{
			}


		/** state (Route_Calc) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (Timer) enter executives **/
		FSM_STATE_ENTER_FORCED (7, state7_enter_exec, "Timer", "DVMRP () [Timer enter execs]")
			{
			int i, j, k;
			ForwardInfo *fwd;
			int ups;
			
			if (OFF) printf("(%d.%d) DVMRP Timer %f\n", net_num0, node_num0, op_sim_time());
			
			i = 0;
			while( ForwardingTable[i].source.NetNum != 0 &&
			       ForwardingTable[i].source.NodeNum != 0 )
			{
				fwd = &ForwardingTable[i];
				
				/* Graft retransmit */
				if ( fwd -> grafttick > 0 )
				{
					fwd -> grafttick--;
					if ( fwd -> grafttick == 0 )
					{
						fwd -> grafttick = GraftResendTimer;
						graft( fwd -> upstream, fwd -> source, fwd -> group );
					}
				}
			
				/* Find how many downstream is currently up */
				ups = 0;
				for( j = 0; j < NoOfLinks; j++ )
				{
					if ( fwd -> downflag == UP )
						ups++;
				}
				/* Prune Timeout */
				for( j = 0; j < NoOfLinks; j++ )
				{
					if ( fwd -> downflag[j] = DOWN && fwd -> downtick[j] > 0 )
					{
						fwd -> downtick[j] --;
						if ( fwd -> downtick[j] == 0 )
						{
							fwd -> downtick[j] = -1;
							fwd -> downflag[j] = UP;
							if ( ups == 0 )
								graft( fwd -> upstream, fwd -> source, fwd -> group );
							ups++;
						}
					}
				}
			
				i++;
			}
			op_intrpt_schedule_self( op_sim_time() + 1, TIMER );
			}


		/** state (Timer) exit executives **/
		FSM_STATE_EXIT_FORCED (7, state7_exit_exec, "Timer", "DVMRP () [Timer exit execs]")
			{
			}


		/** state (Timer) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (Prune) enter executives **/
		FSM_STATE_ENTER_FORCED (8, state8_enter_exec, "Prune", "DVMRP () [Prune enter execs]")
			{
			int group;
			int i, FoundGroup;
			mc_id *mc;
			
			
			/* Only GWY will reach here */
			
			group = *((int*) op_ima_obj_svar_get(igmp_objid, "grp_id"));
			
			VL PRUNE_DEBUG {
			    printf(" (%d.%d) prune state\n", net_num0, node_num0);
			    printf(" group = %d\n", group);
			}
			
			mc = (mc_id *)op_ima_obj_svar_get( ip_objid, "mc_table" );
			
			for( i = 0; i < NoOfLinks; i++ )
			{
				FoundGroup = 0;
			VL PRUNE_DEBUG {
			    printf(" mc_table: (grp %d, interface %d) : %d\n", group, i, mc[group].interface[i]);
			}
			        if( mc[group].interface[i] == 1)
			        {
			       		FoundGroup = 1;
					break; 
				}
			}
			
			/* check forwarding table, send graft to each (source, group) pair */
			if ( FoundGroup == 0 )
			{
			
				i=0;
				while( ForwardingTable[i].source.NetNum != 0 &&
			       		ForwardingTable[i].source.NodeNum != 0 )
				{       
			        	if ( ForwardingTable[i].group == group &&
					     ( ForwardingTable[i].source.NetNum != net_num0 ||
					       ForwardingTable[i].source.NodeNum != node_num0 ) )
			        	{
			                	prune( ForwardingTable[i].upstream,
			                       		ForwardingTable[i].source, group, PruneLifetime );
			                	ForwardingTable[i].downflag[0] == DOWN; 
			                	ForwardingTable[i].downtick[0] = 0;
			                	ForwardingTable[i].grafttick = -1;
			        	}
			        	i++;
				}
			}
			
			}


		/** state (Prune) exit executives **/
		FSM_STATE_EXIT_FORCED (8, state8_exit_exec, "Prune", "DVMRP () [Prune exit execs]")
			{
			}


		/** state (Prune) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,DVMRP)
	}




void
DVMRP_svar (prs_ptr,var_name,var_p_ptr)
	DVMRP_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (DVMRP_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("my_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_my_id);
	if (Vos_String_Equal ("ip_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_objid);
	if (Vos_String_Equal ("igmp_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_igmp_objid);
	if (Vos_String_Equal ("sub_netid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_sub_netid);
	if (Vos_String_Equal ("report_dist" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_report_dist);
	if (Vos_String_Equal ("AreaId" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_AreaId);
	if (Vos_String_Equal ("NoOfLinks" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfLinks);
	if (Vos_String_Equal ("Indep_Rou" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_Indep_Rou);
	if (Vos_String_Equal ("IS_ROUTER" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_IS_ROUTER);
	if (Vos_String_Equal ("IS_DESIGNATED_ROUTER" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_IS_DESIGNATED_ROUTER);
	if (Vos_String_Equal ("IS_AREA_BORDER_ROUTER" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_IS_AREA_BORDER_ROUTER);
	if (Vos_String_Equal ("net_num0" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num0);
	if (Vos_String_Equal ("node_num0" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num0);
	if (Vos_String_Equal ("net_num1" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num1);
	if (Vos_String_Equal ("node_num1" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num1);
	if (Vos_String_Equal ("net_num2" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num2);
	if (Vos_String_Equal ("node_num2" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num2);
	if (Vos_String_Equal ("net_num3" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num3);
	if (Vos_String_Equal ("node_num3" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num3);
	if (Vos_String_Equal ("net_num4" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num4);
	if (Vos_String_Equal ("node_num4" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num4);
	if (Vos_String_Equal ("net_num5" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num5);
	if (Vos_String_Equal ("node_num5" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num5);
	if (Vos_String_Equal ("net_num6" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num6);
	if (Vos_String_Equal ("node_num6" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num6);
	if (Vos_String_Equal ("net_num7" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num7);
	if (Vos_String_Equal ("node_num7" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num7);
	if (Vos_String_Equal ("net_num8" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num8);
	if (Vos_String_Equal ("node_num8" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num8);
	if (Vos_String_Equal ("net_num9" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num9);
	if (Vos_String_Equal ("node_num9" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num9);
	if (Vos_String_Equal ("in_strm" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_in_strm);
	if (Vos_String_Equal ("ip_strm" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_strm);
	if (Vos_String_Equal ("src_net" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_src_net);
	if (Vos_String_Equal ("src_node" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_src_node);
	if (Vos_String_Equal ("grp_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_grp_id);
	if (Vos_String_Equal ("dgram_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_dgram_id);
	if (Vos_String_Equal ("NoOfNeighborElements" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfNeighborElements);
	if (Vos_String_Equal ("InstNoLsa" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_InstNoLsa);
	if (Vos_String_Equal ("LinksInfo" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_LinksInfo);
	if (Vos_String_Equal ("RoutingTable" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_RoutingTable);
	if (Vos_String_Equal ("ForwardingTable" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_ForwardingTable);

	FOUT;
	}




void
DVMRP_diag ()
	{
	int data_len;
	
	int protocolid;
	int pkCode;
	
	Packet *ReceivedDVMRPPkptr;
	Packet *ReceivedPkptr;
	
	ReportStruct *rcvReportPtr;
	ProbeStruct *rcvProbePtr;
	PruneStruct *rcvPrunePtr;
	GraftStruct *rcvGraftPtr;
	
	int _block_origin;

	FIN (DVMRP_diag ())

	BINIT

	FOUT;
	}




void
DVMRP_terminate ()
	{
	int data_len;
	
	int protocolid;
	int pkCode;
	
	Packet *ReceivedDVMRPPkptr;
	Packet *ReceivedPkptr;
	
	ReportStruct *rcvReportPtr;
	ProbeStruct *rcvProbePtr;
	PruneStruct *rcvPrunePtr;
	GraftStruct *rcvGraftPtr;
	
	int _block_origin;

	FIN (DVMRP_terminate ())

	BINIT

	FOUT;
	}



Compcode
DVMRP_init (pr_state_pptr)
	DVMRP_state			**pr_state_pptr;
	{
	int _block_origin = 0;
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	extern int				Vos_Nop ();

	FIN (DVMRP_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (DVMRP)",
			sizeof (DVMRP_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (DVMRP_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



enum { _block_origin = __LINE__ };
void initialize()
{
	int i,j;
	Objid q_id, tx_id, dup_lk_id, rx_id, ip_id;

        VL PROBE_DEBUG printf("Router Id is : %d.%d\n", net_num0, node_num0);

        for(i = 0 ; i < NoOfLinks; i++)
        {
                q_id = op_topo_assoc(ip_objid, OPC_TOPO_ASSOC_OUT, OPC_OBJTYPE_QUEUE, i);
                tx_id = op_topo_assoc( q_id, OPC_TOPO_ASSOC_OUT, OPC_OBJTYPE_PTTX, 0 );
                dup_lk_id = op_topo_assoc( tx_id, OPC_TOPO_ASSOC_OUT, OPC_OBJTYPE_LKSIMP, 0 );
                rx_id = op_topo_assoc( dup_lk_id, OPC_TOPO_ASSOC_OUT, OPC_OBJTYPE_PTRX, 0 );
                ip_id = op_topo_assoc( rx_id, OPC_TOPO_ASSOC_OUT, OPC_OBJTYPE_QUEUE, 0 );

                op_ima_obj_attr_get(ip_id,"net_num0",&LinksInfo[i].NeighborRtrId.NetNum);
                op_ima_obj_attr_get(ip_id,"node_num0",&LinksInfo[i].NeighborRtrId.NodeNum);
		LinksInfo[i].flag = DOWN;
		LinksInfo[i].tick = 0;
                LinksInfo[i].IntfIPAddr.NetNum = net_num0;
                LinksInfo[i].IntfIPAddr.NodeNum = i+1;
/*                switch(i)
                {
                        case 0  : LinksInfo[i].IntfIPAddr.NetNum = net_num0;
                                  LinksInfo[i].IntfIPAddr.NodeNum = node_num0;
                                  break;
                        case 1  : LinksInfo[i].IntfIPAddr.NetNum = net_num1;
                                  LinksInfo[i].IntfIPAddr.NodeNum = node_num1;
				  break;
                        case 2  : LinksInfo[i].IntfIPAddr.NetNum = net_num2;
                                  LinksInfo[i].IntfIPAddr.NodeNum = node_num2;
                                  break;
                        case 3  : LinksInfo[i].IntfIPAddr.NetNum = net_num3;
                                  LinksInfo[i].IntfIPAddr.NodeNum = node_num3;
                                  break;
                        case 4  : LinksInfo[i].IntfIPAddr.NetNum = net_num4;
                                  LinksInfo[i].IntfIPAddr.NodeNum = node_num4;
                                  break;
                        case 7  : LinksInfo[i].IntfIPAddr.NetNum = net_num7;
                                  LinksInfo[i].IntfIPAddr.NodeNum = node_num7;
                                  break;
                        case 8  : LinksInfo[i].IntfIPAddr.NetNum = net_num8;
                                  LinksInfo[i].IntfIPAddr.NodeNum = node_num8;
                                  break;
                        case 9  : LinksInfo[i].IntfIPAddr.NetNum = net_num9;
                                  LinksInfo[i].IntfIPAddr.NodeNum = node_num9;
                                  break;
                        default:  printf("DVMRP Init link index bad %d\n",i);
                                  i=0;
                                  break;
                }*/
        }

	memset( RoutingTable, 0, sizeof(RouteInfo) * MAX_NO_ROUTES );
	RoutingTable[0].source.NetNum = net_num0;
	RoutingTable[0].source.NodeNum = node_num0;
        RoutingTable[0].upstream.NetNum = net_num0;
        RoutingTable[0].upstream.NodeNum = node_num0;
	for( i = 0; i < MAX_NO_LINKS; i++ )
		RoutingTable[0].downstream[i] = -1;
        RoutingTable[0].flag = DOWN;
	RoutingTable[0].metric = 0;
        RoutingTable[0].tick = 0;

}


void probePkPro( ProbeStruct *ps )
{
	int i;
	int FoundSelf = 0;

	for( i = 0; i < ps -> NoOfNeighbors && FoundSelf == 0; i++ )
		if ( ps -> neighbor[i].NetNum == net_num0 &&
		     ps -> neighbor[i].NodeNum == node_num0 )
			FoundSelf = 1;
	
	if ( FoundSelf )
	{
		for( i = 0; i < MAX_NO_NEIGHBORS; i++ )
			if ( LinksInfo[i].NeighborRtrId.NetNum == src_net &&
			     LinksInfo[i].NeighborRtrId.NodeNum == src_node )
			{
				LinksInfo[i].flag = UP;
				LinksInfo[i].tick = 0;
				break;
			}
	}

VL PROBE_DEBUG
{
        printf("--------------  Receive probe from (%d.%d)----------------\n",
		src_net,src_node);
        printf("Router(%d.%d) Link Information \n", net_num0, node_num0 );
        printf("Number of Links is %d\n ", NoOfLinks);
        for( i = 0; i < NoOfLinks; i++ )
        {
                printf("%d Neighbor=%d.%d  ",i,
                        LinksInfo[i].NeighborRtrId.NetNum, LinksInfo[i].NeighborRtrId.NodeNum );
                printf("Interface=%d.%d  ",
                        LinksInfo[i].IntfIPAddr.NetNum, LinksInfo[i].IntfIPAddr.NodeNum );
                printf("flag= %s\n", (LinksInfo[i].flag==UP)? "UP" : "DOWN");
        }
        printf("---------------------------------------------------------\n\n");
}

}


/** 
*** delete downStrm from each item in forwarding table which 'source' is source\
*** check each group, prune messages may need to send.
***/ 
int delDownStrmFwding( IPAddress source, int downStrm )
{
	int i, j, k, groups = 0, DownFound = 0, ups = 0;
        ForwardInfo *fwd;

        for( i = 0; i < MAX_NO_FORWARDS; i++ )
        {
            fwd = &ForwardingTable[i];

            if ( fwd -> source.NetNum == 0 &&
                 fwd -> source.NodeNum == 0 )
                break;

            if ( fwd -> source.NetNum == source.NetNum &&
                 fwd -> source.NodeNum == source.NodeNum )
            {
                groups++;
                ups = 0;
                DownFound = 0;

                /* Check if "downStrm" is in the downlist. */
                for( j = 0; j < NoOfLinks; j++ )
                {
                    if ( fwd -> downstream[j] == -1 )
                        break;

                    if ( fwd -> downstream[j] == downStrm )
                    {
			memcpy( &fwd -> downstream[j],
				&fwd -> downstream[j+1],
				( MAX_NO_LINKS - j - 1 ) * sizeof(int) );
			fwd -> downstream[MAX_NO_LINKS - 1] = -1;
			memcpy( &fwd -> downflag[j],
				&fwd -> downflag[j+1],
				( MAX_NO_LINKS - j - 1 ) * sizeof(RouteStatus) );
			fwd -> downflag[MAX_NO_LINKS - 1] = -1;
			memcpy( &fwd -> downtick[j],
                                &fwd -> downtick[j+1],
                                ( MAX_NO_LINKS - j - 1 ) * sizeof(int) );
                        fwd -> downtick[MAX_NO_LINKS - 1] = -1;

			j--; /* check position j again. */
		    }
		    else if ( fwd -> downflag[j] == UP )
			ups++;
		}

		if ( ups == 0 )
		{
		    prune( fwd -> upstream, fwd -> source, fwd -> group, PruneLifetime );
		} 
	    }
	}
}


/* Add downstream router IP to the forwarding table. The function
   should check all (source, group) pairs. Graft message may need 
   to send. Return how many groups that originate from the source.
*/
int addDownStrmFwding( IPAddress source, int downStrm )
{
	int i, j, k, groups = 0, DownFound = 0, ups = 0;
	ForwardInfo *fwd;

	for( i = 0; i < MAX_NO_FORWARDS; i++ )
	{
	    fwd = &ForwardingTable[i];

	    if ( fwd -> source.NetNum == 0 &&
	         fwd -> source.NodeNum == 0 )
		break;

	    if ( fwd -> source.NetNum == source.NetNum &&
                 fwd -> source.NodeNum == source.NodeNum )
	    {
		groups++;
		ups = 0;
		DownFound = 0;

		/* Check if "downStrm" is in the downlist. */
		for( j = 0; j < NoOfLinks; j++ )
		{
                    if ( fwd -> downstream[j] == -1 )
			break;

		    if ( fwd -> downstream[j] == downStrm ) 
			DownFound = 1;

		    if ( fwd -> downflag[j] == UP )
			ups++;
		}

		if ( DownFound == 0 )
		{
		    fwd -> downstream[j] = downStrm;
		    fwd -> downflag[j] = UP;
		    fwd -> downtick[j] = 0;
		}
		if ( ups == 0 )
		{
			graft( fwd -> upstream, fwd -> source, fwd -> group );
			fwd -> grafttick = GraftResendTimer;
		}
	    }
	}

	return groups;
}




/** Note: 
*** In routing table, we record the IP address of the upstream router, while only
*** record the interface index of the downstream list.
*** All downstream flags will be set to DOWN initially, and will be set to ON after
*** the router gets the poison reverse report or graft.
**/
void reportPkPro( ReportStruct *rs )
{
	int i, j, k;
	int FoundSrc = 0, FoundItem = 0, FoundDown = 0;

VL REPORT_DEBUG
	printf("%d.%d receive report from %d.%d\n", net_num0, node_num0, src_net, src_node);

	/* check if the source is one of my neighbors */
	for( i = 0; i < MAX_NO_NEIGHBORS && FoundSrc == 0; i++ )
        	if ( LinksInfo[i].NeighborRtrId.NetNum == src_net &&
                     LinksInfo[i].NeighborRtrId.NodeNum == src_node &&
		     LinksInfo[i].flag == UP )
			FoundSrc = 1;
	
	if ( FoundSrc == 0 )	return;		
	
	for( i = 0; i < rs -> NoOfRoutes; i++ )
	{
		rs -> metric[i] ++;
		j = 0;
		FoundItem = 0;
		while( RoutingTable[j].source.NetNum != 0 &&
		       RoutingTable[j].source.NodeNum != 0 )
		{
		        if ( RoutingTable[j].source.NetNum == rs -> source[i].NetNum &&
			    RoutingTable[j].source.NodeNum == rs -> source[i].NodeNum )
			{
			/* The routing item already exist */
				FoundItem = 1;

				/* Check if upstream got from report is myself */
				if ( rs -> upstream[i].NetNum == net_num0 &&
				     rs -> upstream[i].NodeNum == node_num0 )
				{
					/* Check if the src of report is one of my downlist */
					k = 0;
					FoundDown = 0;
					while( RoutingTable[j].downstream[k] != -1 &&
					       k < MAX_NO_LINKS	)
					{
						if ( RoutingTable[j].downstream[k] == ip_strm )
						{
							FoundDown = 1;
							/*RoutingTable[j].metric = rs -> metric[i];
							RoutingTable[j].tick = 0;*/
							break;
						}
						k++;
					}
					if ( FoundDown == 0 && k != MAX_NO_LINKS )
					{
						RoutingTable[j].downstream[k] = ip_strm;
						/*RoutingTable[j].metric = rs -> metric[i];*/
                                                RoutingTable[j].tick = 0;

						addDownStrmFwding( rs -> source[i], ip_strm );
					}
				}
				else
				{
					/* delete ip_strm from the downlist */
					k = 0;
					while( RoutingTable[j].downstream[k] != -1 &&
					       k < MAX_NO_LINKS	)
					{
					    if ( RoutingTable[j].downstream[k] == ip_strm )
					    {
						memcpy( &(RoutingTable[j].downstream[k]),
						 	&(RoutingTable[j].downstream[k+1]),
						 	(MAX_NO_LINKS-k-1)*sizeof(int) );
						RoutingTable[j].downstream[MAX_NO_LINKS-1]=-1;

						delDownStrmFwding( rs -> source[i], ip_strm );
						break;
					    }
					    k++;
					}

					if ( rs -> metric[i] > RoutingTable[j].metric )
					{
						/* if the current upstream is the src of report,
						   just refresh routingtable */	
						if ( RoutingTable[j].upstream.NetNum == src_net &&
						     RoutingTable[j].upstream.NodeNum == src_node )
						{
							RoutingTable[j].metric = rs -> metric[i];
							RoutingTable[j].tick = 0;
						}
						else ;
						/* discard report */
					}
					else if ( rs -> metric[i] < RoutingTable[j].metric )
					{
					    RoutingTable[j].metric = rs -> metric[i];
					    if ( RoutingTable[j].upstream.NetNum != src_net ||
				 		 RoutingTable[j].upstream.NodeNum != src_node )
					    {
						RoutingTable[j].upstream.NetNum = src_net;
                                                RoutingTable[j].upstream.NodeNum = src_node;
						RoutingTable[j].tick = 0;
					    }
					}
					else
					{
					    if ( RoutingTable[j].upstream.NetNum == src_net &&
                                                 RoutingTable[j].upstream.NodeNum == src_node )
                             			RoutingTable[j].tick = 0;
					    else if ( src_net < RoutingTable[j].upstream.NetNum )
					    {
                                                RoutingTable[j].upstream.NetNum = src_net;
                                                RoutingTable[j].upstream.NodeNum = src_node;
                                                RoutingTable[j].tick = 0;
                                            }
					    else if ( src_net == RoutingTable[j].upstream.NetNum &&
						      src_node < RoutingTable[j].upstream.NodeNum )
					    {
                                                RoutingTable[j].upstream.NetNum = src_net;          
                                                RoutingTable[j].upstream.NodeNum = src_node;   
                                                RoutingTable[j].tick = 0;      
					    } 
					}	
				}		
			}

			j++;
		}
		if ( FoundItem == 0 &&
		     ( rs -> upstream[i].NetNum != net_num0 ||
		       rs -> upstream[i].NodeNum != node_num0 ) )
		{
			RoutingTable[j].source.NetNum = rs -> source[i].NetNum;
			RoutingTable[j].source.NodeNum = rs -> source[i].NodeNum;
			RoutingTable[j].upstream.NetNum = src_net;
			RoutingTable[j].upstream.NodeNum = src_node;
			RoutingTable[j].metric = rs -> metric[i];
		        for( k = 0; k < MAX_NO_LINKS; k++ )
                		RoutingTable[j].downstream[k] = -1;
			RoutingTable[j].flag = UP;
			RoutingTable[j].tick = 0;
		}
	}

VL REPORT_DEBUG {
    printf("------------- (%d.%d) receive report from (%d.%d) ------------------\n", 
	   net_num0, node_num0, src_net, src_node);
    printf("NoOfRouters = %d\n", rs->NoOfRoutes);
    for (i=0; i< rs->NoOfRoutes; i++) {
        printf("source (%d.%d), upstream (%d.%d) metric %d\n",
                rs->source[i].NetNum, rs->source[i].NodeNum,
                rs->upstream[i].NetNum, rs->upstream[i].NodeNum,
                rs->metric[i]);
    }

    printf("\n\n");
    i=0;
    while( RoutingTable[i].source.NetNum != 0 &&
           RoutingTable[i].source.NodeNum != 0 )
    {
        printf("source (%d.%d), upstream (%d.%d) metric %d downstrm [",
		RoutingTable[i].source.NetNum, RoutingTable[i].source.NodeNum,
		RoutingTable[i].upstream.NetNum, RoutingTable[i].upstream.NodeNum,
		RoutingTable[i].metric);
	for (j=0; j<MAX_NO_LINKS; j++) {
	    printf("%d ",RoutingTable[i].downstream[j]);
	}
	printf("]\n");
        i++;
    }
    printf("total = %d\n",i);
    printf("------------------------------------------------------------\n");
}

}


void prunePkPro( PruneStruct *ps )
{
	int i, j;
	ForwardInfo *fwd;
	int ups;
	mc_id *mc;


VL PRUNE_DEBUG {
    printf("------------- (%d.%d) receive prune from (%d.%d) ------------------\n", 
	   net_num0, node_num0, src_net, src_node);
    printf("Before :\n");
    print_ForwardingTable();
    printf("\n\n");
}

	mc = (mc_id*)op_ima_obj_svar_get( ip_objid, "mc_table" );
	mc[ps -> group].infValid = -1;

	for( i = 0; i < MAX_NO_FORWARDS; i++ )
	{
		fwd = &ForwardingTable[i];

		if ( fwd -> source.NetNum == ps -> source.NetNum &&
		     fwd -> source.NodeNum == ps -> source.NodeNum &&
		     fwd -> group == ps -> group )
		{
			/* check if ip_strm contained in the downlist */
			ups = 0;
			for( j = 0; j < NoOfLinks; j++ )
			{
				if ( fwd -> downstream[j] == ip_strm )
				{
					fwd -> downtick[j] = ps -> lifetime;
					fwd -> downflag[j] = DOWN;
/*      					mc[ps->group].interface[ip_strm] = 0;*/
				}
				
				if ( fwd -> downflag[j] == UP && fwd -> downstream[j] != -1)
					ups ++;
			}
			if ( ups == 0 )
			    prune( fwd -> upstream, fwd -> source, fwd -> group, PruneLifetime );
		}
	}

VL PRUNE_DEBUG {
    printf("After :\n");
    print_ForwardingTable();
    printf("------------------------------------------------------------\n");
}
}


void graftPkPro( GraftStruct *gs )
{
	int i, j;
	ForwardInfo *fwd;
	int ups, FoundDown;
	mc_id *mc;


VL GRAFT_DEBUG {
    printf("------------- (%d.%d) receive graft from (%d.%d) ------------------\n", 
	   net_num0, node_num0, src_net, src_node);
    printf("Before :\n");
    print_ForwardingTable();
    printf("\n\n");
}

	mc = (mc_id*)op_ima_obj_svar_get( ip_objid, "mc_table" );
	mc[gs -> group].infValid = -1;

	for( i = 0; i < MAX_NO_FORWARDS; i++ )
        {
                fwd = &ForwardingTable[i];

                if ( fwd -> source.NetNum == gs -> source.NetNum &&
                     fwd -> source.NodeNum == gs -> source.NodeNum &&
                     fwd -> group == gs -> group )
                {
		/* Send Ack to the srouce of graft;
		   Add downstream if it doesn't exist;
		   Send graft to upstream if necessary.
		*/ 
			graftAck( ip_strm, fwd -> source, fwd -> group );

			ups = 0;
			j = 0;
			FoundDown = 0;
			while( fwd -> downstream[j] != -1 && j < NoOfLinks )
			{
				if ( fwd -> downflag[j] == UP && fwd -> downstream[j] != -1)
					ups++;

				if ( fwd -> downstream[j] == ip_strm )
                                {
                                        fwd -> downtick[j] = -1;
                                        fwd -> downflag[j] = UP;
/*					mc[ps->group].interface[ip_strm] = 1;*/
					FoundDown = 1;
                                }
				j++;
			}

			if ( FoundDown == 0 && j != NoOfLinks )
			{
				fwd -> downstream[j] = ip_strm;
				fwd -> downflag[j] = UP;
				fwd -> downtick[j] = -1;
				/* Do we need update Routing table too? */
			}

			if ( ups == 0 )
			{
				graft( fwd -> upstream, fwd -> source, fwd -> group );
				fwd -> grafttick = GraftResendTimer;
			}
		}
	}
VL GRAFT_DEBUG {
    printf("After :\n");
    print_ForwardingTable();
    printf("------------------------------------------------------------\n");
}
}


void graftAckPkPro( GraftStruct *gs )
{
	int i;
	ForwardInfo *fwd;
	
VL GRAFT_DEBUG {
    printf("------------- (%d.%d) receive graftACK from (%d.%d) ------------------\n", 
	   net_num0, node_num0, src_net, src_node);
    printf("Before :\n");
    print_ForwardingTable();
    printf("\n\n");
}

	for( i = 0; i < MAX_NO_FORWARDS; i++ )
	{
                fwd = &ForwardingTable[i];

                if ( fwd -> source.NetNum == gs -> source.NetNum &&
                     fwd -> source.NodeNum == gs -> source.NodeNum &&
                     fwd -> group == gs -> group )
			fwd -> grafttick = -1;
	}

VL GRAFT_DEBUG {
    printf("After :\n");
    print_ForwardingTable();
    printf("------------------------------------------------------------\n");
}
} 


/**
*** Send a prune message to the upstream router.
**/
void prune( IPAddress dest, IPAddress source, int group, int lifetime )
{
	Packet *prune_pkptr, *ip_pkptr;
	int data_len;
	PruneStruct *ps;
	int i;

        if ( ip2stream( dest)  == -1 ) return;

	prune_pkptr = op_pk_create_fmt( "dvmrp_pkfmt" );
	op_pk_nfd_set( prune_pkptr, "Code", PRUNE_PKCODE );

	ps = (PruneStruct *) op_prg_mem_alloc(sizeof(PruneStruct));
	ps->source.NetNum = source.NetNum;
	ps->source.NodeNum = source.NodeNum;
	ps->group = group;
	ps->lifetime = lifetime;

	op_pk_nfd_set( prune_pkptr ,"Data", ps, op_prg_mem_alloc,
                 op_prg_mem_free, sizeof(PruneStruct) );

	data_len = op_pk_total_size_get( prune_pkptr ) / 8;

	ip_pkptr = op_pk_create_fmt( "ip_dgram" );
	op_pk_nfd_set( ip_pkptr, "data", prune_pkptr );
	op_pk_bulk_size_set( ip_pkptr, data_len * 8 );

	/* Destination address here is the outgoing stream of this node, NOT 
	   the destination IP address. IP layer handle this number seperately. */
	op_pk_nfd_set( ip_pkptr, "dest_net", 0);
	op_pk_nfd_set( ip_pkptr, "dest_node", ip2stream( dest ) );
	op_pk_nfd_set( ip_pkptr, "ttl", 2 );

	op_pk_nfd_set( ip_pkptr, "src_net", net_num0 );
	op_pk_nfd_set( ip_pkptr, "src_node", node_num0 );

	op_pk_nfd_set( ip_pkptr, "orig_len", data_len );
	op_pk_nfd_set( ip_pkptr, "frag_len", data_len );
	op_pk_nfd_set( ip_pkptr, "frag", 0 );

	op_pk_nfd_set( ip_pkptr, "ident", dgram_id++ );
	op_pk_nfd_set( ip_pkptr, "prot_id", DVMRP_PROTOCOL_ID );

	op_pk_send( ip_pkptr, IPSTREAM );
VL PRUNE_DEBUG {
    printf("----------(%d,%d) send prune to (%d.%d)----------\n",
	net_num0, node_num0, 0, ip2stream(dest));
    printf("source (%d.%d) grp (%d) lifetime (%d)\n",
	   ps->source.NetNum, ps->source.NodeNum,
	   ps->group, ps->lifetime);	
    printf("-------------------------------------------------\n");
}
}


void graft( IPAddress dest, IPAddress source, int group )
{
	Packet *graft_pkptr, *ip_pkptr;
	int data_len;
        GraftStruct *ps;
        int i;

        if ( ip2stream( dest)  == -1 ) return;

        graft_pkptr = op_pk_create_fmt( "dvmrp_pkfmt" );
        op_pk_nfd_set( graft_pkptr, "Code", GRAFT_PKCODE );

	ps = (GraftStruct *) op_prg_mem_alloc(sizeof(GraftStruct));
        ps->source.NetNum = source.NetNum;
        ps->source.NodeNum = source.NodeNum;
        ps->group = group;

        op_pk_nfd_set( graft_pkptr ,"Data", ps, op_prg_mem_alloc,
                 op_prg_mem_free, sizeof(GraftStruct) );

        data_len = op_pk_total_size_get( graft_pkptr ) / 8;

        ip_pkptr = op_pk_create_fmt( "ip_dgram" );
        op_pk_nfd_set( ip_pkptr, "data", graft_pkptr );
        op_pk_bulk_size_set( ip_pkptr, data_len * 8 );

	/* Destination address here is the outgoing stream of this node, NOT 
	   the destination IP address. IP layer handle this number seperately. */
        op_pk_nfd_set( ip_pkptr, "dest_net", 0 );
        op_pk_nfd_set( ip_pkptr, "dest_node", ip2stream( dest ) );
        op_pk_nfd_set( ip_pkptr, "ttl", 2 );

        op_pk_nfd_set( ip_pkptr, "src_net", net_num0 );
        op_pk_nfd_set( ip_pkptr, "src_node", node_num0 );

        op_pk_nfd_set( ip_pkptr, "orig_len", data_len );
        op_pk_nfd_set( ip_pkptr, "frag_len", data_len );
        op_pk_nfd_set( ip_pkptr, "frag", 0 );

        op_pk_nfd_set( ip_pkptr, "ident", dgram_id++ );
        op_pk_nfd_set( ip_pkptr, "prot_id", DVMRP_PROTOCOL_ID );

        op_pk_send( ip_pkptr, IPSTREAM );
VL GRAFT_DEBUG {
    printf("----------(%d,%d) send graft to (%d.%d)----------\n",
	net_num0, node_num0, 0, ip2stream(dest));
    printf("source (%d.%d) grp (%d) \n",
	   ps->source.NetNum, ps->source.NodeNum, ps->group);
    printf("-------------------------------------------------\n");
}
}


void graftAck( int dest, IPAddress source, int group )
{
       
	Packet *graft_pkptr, *ip_pkptr;
	int data_len;
        GraftStruct *ps;
        int i;

	graft_pkptr = op_pk_create_fmt( "dvmrp_pkfmt" );
        op_pk_nfd_set( graft_pkptr, "Code", GRAFTACK_PKCODE );

   	ps = (GraftStruct *)op_prg_mem_alloc(sizeof(GraftStruct));
        ps->source.NetNum = source.NetNum;
        ps->source.NodeNum = source.NodeNum;
        ps->group = group;


        op_pk_nfd_set( graft_pkptr ,"Data", ps, op_prg_mem_alloc,
                 op_prg_mem_free, sizeof(GraftStruct) );

        data_len = op_pk_total_size_get( graft_pkptr ) / 8;

        ip_pkptr = op_pk_create_fmt( "ip_dgram" );
        op_pk_nfd_set( ip_pkptr, "data", graft_pkptr );
        op_pk_bulk_size_set( ip_pkptr, data_len * 8 );
        
	/* Destination address here is the outgoing stream of this node, NOT 
	   the destination IP address. IP layer handle this number seperately. */
	op_pk_nfd_set( ip_pkptr, "dest_net", 0 ); 
        op_pk_nfd_set( ip_pkptr, "dest_node", dest);
        op_pk_nfd_set( ip_pkptr, "ttl", 2 );

        op_pk_nfd_set( ip_pkptr, "src_net", net_num0 );
        op_pk_nfd_set( ip_pkptr, "src_node", node_num0 );

        op_pk_nfd_set( ip_pkptr, "orig_len", data_len );
        op_pk_nfd_set( ip_pkptr, "frag_len", data_len );
        op_pk_nfd_set( ip_pkptr, "frag", 0 );

        op_pk_nfd_set( ip_pkptr, "ident", dgram_id++ );
        op_pk_nfd_set( ip_pkptr, "prot_id", DVMRP_PROTOCOL_ID );

        op_pk_send( ip_pkptr, IPSTREAM );
VL GRAFT_DEBUG {
    printf("----------(%d,%d) send graftACK to (%d.%d)----------\n",
	net_num0, node_num0, 0, dest);
    printf("source (%d.%d) grp (%d)\n",
	   ps->source.NetNum, ps->source.NodeNum, ps->group);
    printf("----------------------------------------------------\n");
}
}



int ip2stream( IPAddress upstrm )
{

	int i;


	for ( i = 0; i < NoOfLinks; i++ )
	{
		if ( LinksInfo[i].NeighborRtrId.NetNum == upstrm.NetNum &&
		     LinksInfo[i].NeighborRtrId.NodeNum == upstrm.NodeNum ) {
if (OFF) printf("IP2stream : from (%d.%d) to (%d.%d)\n", 
	upstrm.NetNum, upstrm.NodeNum,
	LinksInfo[i].IntfIPAddr.NetNum,LinksInfo[i].IntfIPAddr.NodeNum);
			return LinksInfo[i].IntfIPAddr.NodeNum;
		}
	}
	return -1;
}

void print_ForwardingTable() {
    int i=0, j;

    while( ForwardingTable[i].source.NetNum != 0 &&
           ForwardingTable[i].source.NodeNum != 0 )
    {
        printf("source (%d.%d), GRP %d upstream (%d.%d) metric %d gtick %d \n\tdownstrm [",
		ForwardingTable[i].source.NetNum, ForwardingTable[i].source.NodeNum,
		ForwardingTable[i].group,
		ForwardingTable[i].upstream.NetNum, ForwardingTable[i].upstream.NodeNum,
		ForwardingTable[i].metric,
		ForwardingTable[i].grafttick);
	for (j=0; j<MAX_NO_LINKS; j++) {
	    printf("(%d %s %d)", ForwardingTable[i].downstream[j],
		(ForwardingTable[i].downflag[j]==UP)? "UP":"DOWN",
		ForwardingTable[i].downtick[j]);
	}
	printf("]\n");
        i++;
    }
    printf("total = %d\n",i);
}
