/* Process model C form file: ip_host_rte.pr.c */
/* Portions of this file copyright 1992, 1996 by MIL 3, Inc. */



/* OPNET system definitions */
#include <opnet.h>
#include "/disk2/gduan/qosip/host_ip/ip_host_rte.pr.h"
FSM_EXT_DECS



/* Header block */
#define TRACE 	if(0)

#include "ip_frag_sup.h"

/* 
** Macros Used to Represent Transitions. 
*/
#define QUEUE_EMPTY     	(op_q_empty ())
#define SVC_COMPLETION  	op_intrpt_type () == OPC_INTRPT_SELF
#define ARRIVAL         	op_intrpt_type () == OPC_INTRPT_STRM
#define RSVP 46
#define FD_GWY_ND 2

/*
** Protocol ID's and Versions of RSVP and OSPF protocols simulated
*/
#define IGMP_PROTOCOL_ID	9
#define TCP_PROTOCOL_ID		12
#define QOSPF_PROTOCOL_ID	89
#define RSVP_PROTOCOL_ID	46
#define UDP_PROTOCOL_ID		20
#define RSVP_VERSION		1
#define OSPF_VERSION		1


/*
** Streams of the model. Packets are sent on these streams 
*/
#define IGMP_STREAM 		0
#define ARP_STREAM 			1
#define DEF_STREAM			1
#define RAPP_STREAM			2
#define RSVP_STREAM			3
#define UAPP_STREAM			4
#define MAPP_STREAM			5

#define VL 				if(OFF)
#define MCVL				if(OFF)
#define IDLE 				0
#define BUSY				1
#define OFF					0
#define ON					1
#define FALSE				0
#define TRUE				1
#define MTU                 4096
#define UNDEFINED			-1
#define MAX_ROU_TABLE_SIZE  5
/* 
** Default value for the time-to-live counter. This value is assigned when a packet is received 
** which does not already have a ttl field. 
*/
#define IP_DEFAULT_TTL			32
#define IS_MULTICAST			500

ip_dgram_list* ip_frag_sup_setup();

/*
** IP Address structure
*/
typedef struct 
{
    int NetNum;
    int NodeNum;
} IPAddress;

#define TOTAL_DESTROY_STAT 10
double total_destroy;
int id;

/* global statictics variable*/
double g_host_ip_pkt_drop;
#if !defined (VOSD_NO_FIN)
#undef	BIN
#undef	BOUT
#define	BIN		_fstack_local_info.last_line_passed = __LINE__ - _block_origin;
#define	BOUT	BIN
#define	BINIT	_fstack_local_info.last_line_passed = 0; _block_origin = __LINE__;
#else
#define	BINIT
#endif /* #if !defined (VOSD_NO_FIN) */



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	int                     		sv_server_busy;
	int                     		sv_def_net;
	int                     		sv_def_nod;
	double                  		sv_service_rate;
	Objid                   		sv_own_id;
	int	                    		sv_net_num0;
	int	                    		sv_node_num0;
	int	                    		sv_net_num1;
	int	                    		sv_node_num1;
	int	                    		sv_net_num2;
	int	                    		sv_node_num2;
	Ici*	                   		sv_arp_iciptr;
	ip_dgram_list*	         		sv_dgram_list_ptr;
	int	                    		sv_default_ttl;
	int	                    		sv_rte_tab[500];
	Stathandle              		sv_g_host_ip_pkt_drop_handler;
	} ip_host_rte_state;

#define pr_state_ptr            		((ip_host_rte_state*) SimI_Mod_State_Ptr)
#define server_busy             		pr_state_ptr->sv_server_busy
#define def_net                 		pr_state_ptr->sv_def_net
#define def_nod                 		pr_state_ptr->sv_def_nod
#define service_rate            		pr_state_ptr->sv_service_rate
#define own_id                  		pr_state_ptr->sv_own_id
#define net_num0                		pr_state_ptr->sv_net_num0
#define node_num0               		pr_state_ptr->sv_node_num0
#define net_num1                		pr_state_ptr->sv_net_num1
#define node_num1               		pr_state_ptr->sv_node_num1
#define net_num2                		pr_state_ptr->sv_net_num2
#define node_num2               		pr_state_ptr->sv_node_num2
#define arp_iciptr              		pr_state_ptr->sv_arp_iciptr
#define dgram_list_ptr          		pr_state_ptr->sv_dgram_list_ptr
#define default_ttl             		pr_state_ptr->sv_default_ttl
#define rte_tab                 		pr_state_ptr->sv_rte_tab
#define g_host_ip_pkt_drop_handler		pr_state_ptr->sv_g_host_ip_pkt_drop_handler

/* This macro definition will define a local variable called	*/
/* "op_sv_ptr" in each function containing a FIN statement.	*/
/* This variable points to the state variable data structure,	*/
/* and can be used from a C debugger to display their values.	*/
#undef FIN_PREAMBLE
#define FIN_PREAMBLE	ip_host_rte_state *op_sv_ptr = pr_state_ptr;




/* Process model interrupt handling procedure */



void
ip_host_rte ()
	{
	double          pk_svc_time;
	int             insert_ok,protocol;
	int				in_strm,outstrm, mtu, frag;
	int				next_net, next_node;
	int				dest_net, dest_node,src_net,src_node;
	int				i, frag_accum, num_frags;
	int				frag_size, data_size, len;
	int				header_size, ttl;
	Packet			*pkptr, *ip_pkptr, *frag_ptr;
	Packet			*data_pkptr;
	Packet 			*rsvp_pkptr, *tmp_pkptr;
	int _block_origin = 0;


	FSM_ENTER (ip_host_rte)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (init) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "init", "ip_host_rte () [init enter execs]")
			{
			/****************************************************************************************************
			** Process Model : ip_host_rte                                                                     **
			** State         : init                                                                            **
			** Authors       : Ravi Malghan(rmalghan@descartes.gmu.edu) & Lava K Lavu(llavu@nac.gmu.edu)       **
			**                                                                                                 **
			** This state intializes all the variables that are used in the module, create an datagram list to **
			** hold the fragments, get all the attributes from the configuration file                          **
			**                                                                                                 **
			****************************************************************************************************/
			
			server_busy = IDLE;        /* Intialize the server to be idle */
			
			/* 
			** Get the object id of this module and load all the attributes from the configuration file
			*/
			own_id = op_id_self ();
			op_ima_obj_attr_get (own_id, "service_rate", &service_rate);   /*     Service rate attribute       */
			op_ima_obj_attr_get (own_id, "net_num0", &net_num0);           /*     Following 6 lines are the    */
			op_ima_obj_attr_get (own_id, "net_num1", &net_num1);		   /*   Network Addresses of Various   */
			op_ima_obj_attr_get (own_id, "net_num2", &net_num2);           /*            interfaces            */
			op_ima_obj_attr_get (own_id, "node_num0", &node_num0);
			op_ima_obj_attr_get (own_id, "node_num1", &node_num1);
			op_ima_obj_attr_get (own_id, "node_num2", &node_num2);
			
			/*Added by Jiemei */
			TRACE printf("HostIp: MyobjId %d  service_rate %f  net_num %d %d %d  node_num %d %d %d\n",own_id,service_rate,net_num0,net_num1,net_num2,node_num0,node_num1,node_num2);
			service_rate=520;
			
			/* 
			** Create a datagram list to hold datagram fragments until the complete datagram can be 
			** reconstructed from the fragments. 
			*/
			dgram_list_ptr = ip_frag_sup_setup ();
			
			arp_iciptr = op_ici_create("ip_arp_req");      /* ICI to communicate the destination address to arp */
			
			for( i = 0; i < MAX_ROU_TABLE_SIZE; i++ )
			/*	rte_tab[i] = 0;*/            /* Intialize the routing table intially to have no predefined routes */
				rte_tab[i]=UNDEFINED;  /*changed by Jiemei*/
			
			total_destroy=0; /*added by Jiemei*/
			
			/*statistic init*/
			g_host_ip_pkt_drop_handler = op_stat_reg("G host ip pkt drop (pkts)", OPC_STAT_INDEX_NONE, OPC_STAT_GLOBAL);
			g_host_ip_pkt_drop = 0;
			}


		/** state (init) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "init", "ip_host_rte () [init exit execs]")
			{
			}


		/** state (init) transition processing **/
		FSM_INIT_COND (ARRIVAL)
		FSM_DFLT_COND
		FSM_TEST_LOGIC ("init")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 1, state1_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 2, state2_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (arrival) enter executives **/
		FSM_STATE_ENTER_FORCED (1, state1_enter_exec, "arrival", "ip_host_rte () [arrival enter execs]")
			{
			/**********************************************************************************************************
			** Process Model : ip_host_rte                                                                           **
			** State         : arrival                                                                               **
			**                                                                                                       **
			** The transition to  this state occurs when there is packet arrival to this module  form any of one its **
			** streams. This state checks for the ttl field and if it is not set then sets that field to the default **
			** ttl value(32). It enqueues the packet into a subqueue for service                                     **
			**                                                                                                       **
			**********************************************************************************************************/
			
			
			/* 
			** acquire the arriving packet. Multiple arriving streams are supported. 
			*/
			in_strm = op_intrpt_strm ();
			pkptr = op_pk_get (in_strm);
			op25_pk_nfd_set(pkptr,"strm",in_strm);
			
			
			/* 
			** If the received packet does not yet have a time-to-live field assignment, place the default value in 
			** this  field. In this  model, ttl is  treated as a  hop allowance  decremented  by one prior  to each 
			** transmission. If the decremented counter reaches zero, the packet is not transmitted. 
			*/
			
			if( !op_pk_nfd_is_set (pkptr, "ttl") )
			{
			    op25_pk_nfd_set (pkptr, "ttl", IP_DEFAULT_TTL);
			}
			/* 
			** attempt to enqueue the packet at tail of subqueue 0 
			*/
			if (op_subq_pk_insert (0, pkptr, OPC_QPOS_TAIL) != OPC_QINS_OK)
			{
				/* 
				** If insertion fails(due to a full queue), deallocate the packet set flag indicating that insertion 
				** failed. This flag is used to determine the transition from this state  */
				
			    /*added by Jiemei*/
			    total_destroy++;
			    op_stat_local_write(TOTAL_DESTROY_STAT,total_destroy);
			
			    g_host_ip_pkt_drop ++;
			    op_stat_write(g_host_ip_pkt_drop_handler, g_host_ip_pkt_drop);
			
			    op_pk_destroy(pkptr);           
			    insert_ok = FALSE;		            
			}
			else
			{
			    insert_ok = TRUE;
			}
			}


		/** state (arrival) exit executives **/
		FSM_STATE_EXIT_FORCED (1, state1_exit_exec, "arrival", "ip_host_rte () [arrival exit execs]")
			{
			}


		/** state (arrival) transition processing **/
		FSM_INIT_COND (!server_busy && insert_ok)
		FSM_DFLT_COND
		FSM_TEST_LOGIC ("arrival")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 3, state3_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 2, state2_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (idle) enter executives **/
		FSM_STATE_ENTER_UNFORCED (2, state2_enter_exec, "idle", "ip_host_rte () [idle enter execs]")
			{
			
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (5,ip_host_rte)


		/** state (idle) exit executives **/
		FSM_STATE_EXIT_UNFORCED (2, state2_exit_exec, "idle", "ip_host_rte () [idle exit execs]")
			{
			}


		/** state (idle) transition processing **/
		FSM_INIT_COND (ARRIVAL)
		FSM_TEST_COND (SVC_COMPLETION)
		FSM_TEST_LOGIC ("idle")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 1, state1_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 4, state4_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (svc_start) enter executives **/
		FSM_STATE_ENTER_FORCED (3, state3_enter_exec, "svc_start", "ip_host_rte () [svc_start enter execs]")
			{
			/**********************************************************************************************************
			** Process Model : ip_host_rte                                                                           **
			** State         : svc_start                                                                             **
			**                                                                                                       **
			** Transition to this state is from the arrival state and the transition condition is a packet should be **
			** inserted into the sub queue. This state on default transits to the idle state.                        **
			**                                                                                                       **
			**********************************************************************************************************/
			
			pk_svc_time = 1.0/service_rate;    /* Time required to service each packet */
			/* 
			** schedule an interrupt for this process at the time where service ends. 
			*/
			op_intrpt_schedule_self (op_sim_time () + pk_svc_time, 0);
			server_busy = BUSY;
			}


		/** state (svc_start) exit executives **/
		FSM_STATE_EXIT_FORCED (3, state3_exit_exec, "svc_start", "ip_host_rte () [svc_start exit execs]")
			{
			}


		/** state (svc_start) transition processing **/
		FSM_TRANSIT_FORCE (2, state2_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (svc_compl) enter executives **/
		FSM_STATE_ENTER_FORCED (4, state4_enter_exec, "svc_compl", "ip_host_rte () [svc_compl enter execs]")
			{
			/******************************************************************************************************
			** Process Module : ip_host_rte                                                                      **
			** State          : svc_compl                                                                        **
			**                                                                                                   **
			** This state gets the packet from the queue. If  the packets have reached their destination then it **
			** reassembles the packet before sending it to the application layer. If it hasn't reached the final **
			** destination it sends that packet to next hop. It also performs fragmentation of the packet if the **
			** data length is more than mtu(maximum transmission unit) This state on transits to svc_stat if the **
			** queue is not empty. It transits to idle state on default(if queue is empty)                       **
			**                                                                                                   **
			******************************************************************************************************/
			
			/* 
			** extract packet at head of queue this is the packet just finishing service 
			*/
			pkptr = op_subq_pk_remove (0, OPC_QPOS_HEAD);
			
			/* 
			** An IP packet has completed service. Obtain the destination network and node numbers. 
			*/
			op_pk_nfd_get (pkptr, "dest_node", &dest_node);
			op_pk_nfd_get (pkptr, "dest_net", &dest_net);
			op_pk_nfd_get (pkptr, "src_net", &src_net);
			op_pk_nfd_get (pkptr, "src_node", &src_node);
			op_pk_nfd_get (pkptr, "strm", &in_strm);
			op_pk_nfd_get (pkptr, "prot_id", &protocol);
			
			/*
			** If the packet is not destinated to this host and if the packet is not a multicast packet and 
			** if there is no  route to the destination in  this host's routing table then drop  the packet 
			*/
			/*if((dest_net != net_num0 && dest_node != node_num0) && dest_net != IS_MULTICAST && 
				rte_tab[dest_net] == UNDEFINED)*/
			if (!(src_net==net_num0 && src_node==node_num0) && !(dest_net==net_num0 && dest_node==node_num0) &&
				dest_net != IS_MULTICAST && rte_tab[dest_net] ==UNDEFINED)  /*changed by Jiemei, Mar 28, 97*/
			{ 
			/* 	op_pk_print(pkptr);*/
			        
			        /*added by Jiemei*/
			        total_destroy++;
			        op_stat_local_write(TOTAL_DESTROY_STAT,total_destroy);
				
				op_pk_destroy(pkptr);
			
				/*comment and change by Jiemei, Mar 28, 97*/
				/*printf("Destroying pk b'cos no route at %d.%d\n",net_num0,node_num0);*/
				printf("Destroy PK from %d.%d to %d.%d (protocol %d) misdirected to host %d.%d\n",src_net,src_node,dest_net,dest_node,protocol,net_num0,node_num0);
				
			}
			else
			{
			
				/* 
				** If the packets have reached their destination, reassembly of ip fragments is done. 
				*/
				if( (dest_net == net_num0 && dest_node == node_num0) ||
					(dest_net == net_num1 && dest_node == node_num1) ||
					(dest_net == net_num2 && dest_node == node_num2) ||
					(dest_net == IS_MULTICAST && !( in_strm == IGMP_STREAM || in_strm == MAPP_STREAM ||
					 in_strm == RSVP_STREAM || in_strm == RAPP_STREAM )) )
				{
					/*
					**  Call the fragment reassembly procedure to determine if a complete packet has arrived. 
					**  Note that the packet will be destroyed in this procedure call unless it is itself the 
					**  complete datagram (i.e., it isn't a fragment) 
					*/
					if( (ip_pkptr = ip_frag_sup_insert (dgram_list_ptr, pkptr)) != OPC_NIL )
					{
						if(protocol == IGMP_PROTOCOL_ID) 
							outstrm = IGMP_STREAM;
						if(protocol == UDP_PROTOCOL_ID) 
							outstrm = UAPP_STREAM;
						if(protocol == RSVP_PROTOCOL_ID) 
							outstrm =RSVP_STREAM;
			
						if(src_net != net_num0 || src_node != node_num0) 
							op_pk_send (ip_pkptr, outstrm);         /* Send the packet to the higher layer. */
						else
							op_pk_destroy(ip_pkptr);
			
			/*			if(src_net != net_num0 || src_node != node_num0)
							op_pk_send (ip_pkptr, outstrm);  */       /* Send the packet to the higher layer. */
			/*			else
							op_pk_destroy(ip_pkptr); */
					}
				}
				/* 
				** For packets which have not reached their final destination, they are send to their next hop. 
				*/
				else
				{ 
					outstrm = DEF_STREAM;
					mtu = MTU;
			
					/* 
					** Decrement the packet's time-to-live field. If zero is reached, discard the packet rather 
					** than send it on. 
					*/
					if(dest_net == IS_MULTICAST) 
						next_node = dest_net +dest_node;
					else 
					{
						if(dest_net == net_num0) 
							next_node = dest_node;
						else 
							next_node = FD_GWY_ND;	/* ??? FD_GWY_ND is 2, DG, 9/25/97 */
					}
					op_pk_nfd_get (pkptr, "ttl", &ttl);
			
					if( dest_net == IS_MULTICAST )
						ttl--;
					if( ttl == 0 )
					{
				                /*added by Jiemei*/
			                        total_destroy++;
			                        op_stat_local_write(TOTAL_DESTROY_STAT,total_destroy);	
			                  	
						op_pk_destroy (pkptr);               /* Destroy the packet due to an expired ttl.       */
					}
					else
					{
						op25_pk_nfd_set (pkptr, "ttl", ttl);   /* Assign the new decremented value of ttl.        */
			
			
						/* 
						** Install an Ici indicating to the lower layer what the address of the next (intermediate) 
						** node is. 
						*/
						op_ici_attr_set (arp_iciptr, "next_node", next_node);
						op_ici_install (arp_iciptr);
			
						frag_size = op_pk_total_size_get(pkptr)/8;	 /* Obtain the size of fragment in bytes     */
						op_pk_nfd_get(pkptr, "frag_len", &data_size);/* Number of bytes of data in this fragment */
			
						header_size = frag_size - data_size;         /* Size of Header                           */
						/* 
						** If the fragment is smaller than the maximum transfer unit, send it as is. 
						*/
						if (frag_size <= mtu)
						{
							op_pk_send (pkptr, outstrm);
						}
						else
						{			
							/* 
							** Otherwise, break it into fragments. Each fragment can contain up to 
							** (mtu - header_size) bytes of data 
							*/
							num_frags = (data_size + mtu - header_size - 1) / (mtu - header_size);
			
			
							/* 
							** If  the fragment is carrying the original datagram  given to IP, extract it 
							** before copies are made. Only one fragment can carry the original packet for 
							** the reassembly model to work properly. 
							*/
							if( op_pk_nfd_is_set( pkptr, "ip_dgram" ) )
							op_pk_nfd_get( pkptr, "ip_dgram", &ip_pkptr );
							else ip_pkptr = OPC_NIL;
			
							/* 
							** If the packet is carrying any encapsulated data (normally this would happen 
							** only  for a packet fragmented for the first time), extract this data packet 
							** so that it will not appear in each fragment generated by copying. 
							*/
							if( op_pk_nfd_is_set( pkptr, "data" ) )
								op_pk_nfd_get( pkptr, "data", &data_pkptr );
							else data_pkptr = OPC_NIL;
			
							for (frag_accum = 0, i = 0; i < num_frags; i++)  /* Create the fragments.         */
							{
								frag_ptr = op_pk_copy (pkptr);        /* Make a copy of the original packet.  */
								op25_pk_nfd_set (frag_ptr, "frag", 1);  /* Indicate that the copy is a fragment */
			
								/* 
								** for all but the last fragment, the size is the mtu and the encapsulated ip 
								** packet is not included. 
								*/
								if (i < num_frags - 1)
								{
									op25_pk_nfd_set (frag_ptr, "frag_len", mtu - header_size);
									op_pk_total_size_set (frag_ptr, 8 * mtu);
									frag_accum += (mtu - header_size);  
								}
								else
								{
									len = data_size - frag_accum;
									op25_pk_nfd_set( frag_ptr, "frag_len", len );
									op_pk_total_size_set( frag_ptr,	8*(header_size + len) );
			
									/* 
									** If the original packet was not a fragment, encapsulate it into the last 
									** fragment created here. 
									*/
									op_pk_nfd_get( pkptr, "frag", &frag );
									if( !frag )
									{
										/* 
										** If the packet  contained encapsulated data (i.e., from the transport), 
										** that  data would  have been removed to  avoid its duplication  in the 
										** fragments. The data should now be reinserted into the original packet. 
										*/
										if( data_pkptr != OPC_NIL )
											op25_pk_nfd_set( pkptr, "data", data_pkptr );
			
										/* In either case the original packet is encapsulated in the fragment. */
										op25_pk_nfd_set( frag_ptr, "ip_dgram", pkptr );
									}
									else /* Otherwise the packet can be discarded. */
									{
										op_pk_destroy (pkptr);
			
										/* 
										** Also, if the packet includes the  original  datagram from which it was 
										** generated, transfer that datagram into the last fragment created here. 
										** Note that it is possible, in the case where a fragment is itself being 
										** fragmented, none of  the  created fragments  will contain the original 
										** datagram. 
										*/
										if (ip_pkptr != OPC_NIL)
										{
											op25_pk_nfd_set (frag_ptr, "ip_dgram", ip_pkptr);
										}
									}
								}
								op_pk_send (frag_ptr, outstrm);          /* Forward the datagram fragment.      */
							}	
						}
					}
				}
			}
			
			
			server_busy = IDLE;                 /*        The IP routing server becomes available again.        */
			}


		/** state (svc_compl) exit executives **/
		FSM_STATE_EXIT_FORCED (4, state4_exit_exec, "svc_compl", "ip_host_rte () [svc_compl exit execs]")
			{
			}


		/** state (svc_compl) transition processing **/
		FSM_INIT_COND (!QUEUE_EMPTY)
		FSM_DFLT_COND
		FSM_TEST_LOGIC ("svc_compl")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 3, state3_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 2, state2_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,ip_host_rte)
	}




void
ip_host_rte_svar (prs_ptr,var_name,var_p_ptr)
	ip_host_rte_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (ip_host_rte_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("server_busy" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_server_busy);
	if (Vos_String_Equal ("def_net" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_def_net);
	if (Vos_String_Equal ("def_nod" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_def_nod);
	if (Vos_String_Equal ("service_rate" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_service_rate);
	if (Vos_String_Equal ("own_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_own_id);
	if (Vos_String_Equal ("net_num0" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num0);
	if (Vos_String_Equal ("node_num0" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num0);
	if (Vos_String_Equal ("net_num1" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num1);
	if (Vos_String_Equal ("node_num1" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num1);
	if (Vos_String_Equal ("net_num2" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num2);
	if (Vos_String_Equal ("node_num2" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num2);
	if (Vos_String_Equal ("arp_iciptr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_arp_iciptr);
	if (Vos_String_Equal ("dgram_list_ptr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_dgram_list_ptr);
	if (Vos_String_Equal ("default_ttl" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_default_ttl);
	if (Vos_String_Equal ("rte_tab" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_rte_tab);
	if (Vos_String_Equal ("g_host_ip_pkt_drop_handler" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_g_host_ip_pkt_drop_handler);

	FOUT;
	}




void
ip_host_rte_diag ()
	{
	double          pk_svc_time;
	int             insert_ok,protocol;
	int				in_strm,outstrm, mtu, frag;
	int				next_net, next_node;
	int				dest_net, dest_node,src_net,src_node;
	int				i, frag_accum, num_frags;
	int				frag_size, data_size, len;
	int				header_size, ttl;
	Packet			*pkptr, *ip_pkptr, *frag_ptr;
	Packet			*data_pkptr;
	Packet 			*rsvp_pkptr, *tmp_pkptr;
	int _block_origin;

	FIN (ip_host_rte_diag ())

	BINIT
	op_prg_odb_print_major ("State of Reassembly Table:", OPC_NIL);
	ip_frag_sup_print (dgram_list_ptr);

	FOUT;
	}




void
ip_host_rte_terminate ()
	{
	double          pk_svc_time;
	int             insert_ok,protocol;
	int				in_strm,outstrm, mtu, frag;
	int				next_net, next_node;
	int				dest_net, dest_node,src_net,src_node;
	int				i, frag_accum, num_frags;
	int				frag_size, data_size, len;
	int				header_size, ttl;
	Packet			*pkptr, *ip_pkptr, *frag_ptr;
	Packet			*data_pkptr;
	Packet 			*rsvp_pkptr, *tmp_pkptr;
	int _block_origin;

	FIN (ip_host_rte_terminate ())

	BINIT

	FOUT;
	}



Compcode
ip_host_rte_init (pr_state_pptr)
	ip_host_rte_state			**pr_state_pptr;
	{
	int _block_origin = 0;
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	extern int				Vos_Nop ();

	FIN (ip_host_rte_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (ip_host_rte)",
			sizeof (ip_host_rte_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (ip_host_rte_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



enum { _block_origin = __LINE__ };

