/* Process model C form file: UniApp.pr.c */
/* Portions of this file copyright 1992, 1996 by MIL 3, Inc. */



/* OPNET system definitions */
#include <opnet.h>
#include "/disk2/gduan/qosip/host_uapp/UniApp.pr.h"
FSM_EXT_DECS



/* Header block */
#define	TRACE	if(0)

/*
** This are the conditions for transitions to got to different states from the idle block
*/
#define	PK_ARR		op_intrpt_type () == OPC_INTRPT_STRM
#define	SELF_INT1	op_intrpt_type () == OPC_INTRPT_SELF  && op_intrpt_code () == 0
#define END_SIM		op_intrpt_type () ==  OPC_INTRPT_ENDSIM

/*
** Different probes that are used to meausure the number of data packets sent, no. of data
** packets received by this module (or process)
*/
#define send_data_intr      0
#define IP_STREAM           0

#define Meas_Intvl          1

#define UDP_PROTOCOL_ID     20

#define ON                  1
#define OFF                 0

#define VL                  if(ON)
#define UADL                if(OFF)
#define UAREVDL             if(OFF)

#define PROBE_ON            if(ON)
#define PROBE_OFF           if(OFF)

typedef struct 
{
	int NetNum;
	int NodeNum;
}IPAddress;

typedef struct 
{
	char Message[500];
}App_Data;

/* global statistic variable */
double g_pkt_recv;
double g_bit_recv;
double g_delay;
double g_avg_hops;

#if !defined (VOSD_NO_FIN)
#undef	BIN
#undef	BOUT
#define	BIN		_fstack_local_info.last_line_passed = __LINE__ - _block_origin;
#define	BOUT	BIN
#define	BINIT	_fstack_local_info.last_line_passed = 0; _block_origin = __LINE__;
#else
#define	BINIT
#endif /* #if !defined (VOSD_NO_FIN) */



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	Objid                   		sv_my_objid;
	Objid                   		sv_ip_objid;
	double                  		sv_data_iat;
	double                  		sv_print_time;
	int                     		sv_MAX_NO_NETS;
	int                     		sv_MAX_NO_NODES;
	int                     		sv_DataSendInitTime;
	Distribution*           		sv_net_dist;
	Distribution*           		sv_node_dist;
	IPAddress               		sv_LocalAddress;
	Stathandle	             		sv_l_pkt_recv_handler;
	Stathandle	             		sv_l_bit_recv_handler;
	Stathandle	             		sv_l_delay_handler;
	Stathandle	             		sv_g_pkt_recv_handler;
	Stathandle	             		sv_g_bit_recv_handler;
	Stathandle	             		sv_g_delay_handler;
	Stathandle	             		sv_g_avg_hops_handler;
	double                  		sv_l_pkt_recv;
	double                  		sv_l_bit_recv;
	double                  		sv_l_delay;
	} UniApp_state;

#define pr_state_ptr            		((UniApp_state*) SimI_Mod_State_Ptr)
#define my_objid                		pr_state_ptr->sv_my_objid
#define ip_objid                		pr_state_ptr->sv_ip_objid
#define data_iat                		pr_state_ptr->sv_data_iat
#define print_time              		pr_state_ptr->sv_print_time
#define MAX_NO_NETS             		pr_state_ptr->sv_MAX_NO_NETS
#define MAX_NO_NODES            		pr_state_ptr->sv_MAX_NO_NODES
#define DataSendInitTime        		pr_state_ptr->sv_DataSendInitTime
#define net_dist                		pr_state_ptr->sv_net_dist
#define node_dist               		pr_state_ptr->sv_node_dist
#define LocalAddress            		pr_state_ptr->sv_LocalAddress
#define l_pkt_recv_handler      		pr_state_ptr->sv_l_pkt_recv_handler
#define l_bit_recv_handler      		pr_state_ptr->sv_l_bit_recv_handler
#define l_delay_handler         		pr_state_ptr->sv_l_delay_handler
#define g_pkt_recv_handler      		pr_state_ptr->sv_g_pkt_recv_handler
#define g_bit_recv_handler      		pr_state_ptr->sv_g_bit_recv_handler
#define g_delay_handler         		pr_state_ptr->sv_g_delay_handler
#define g_avg_hops_handler      		pr_state_ptr->sv_g_avg_hops_handler
#define l_pkt_recv              		pr_state_ptr->sv_l_pkt_recv
#define l_bit_recv              		pr_state_ptr->sv_l_bit_recv
#define l_delay                 		pr_state_ptr->sv_l_delay

/* This macro definition will define a local variable called	*/
/* "op_sv_ptr" in each function containing a FIN statement.	*/
/* This variable points to the state variable data structure,	*/
/* and can be used from a C debugger to display their values.	*/
#undef FIN_PREAMBLE
#define FIN_PREAMBLE	UniApp_state *op_sv_ptr = pr_state_ptr;




/* Process model interrupt handling procedure */



void
UniApp ()
	{
	double thru, thru1, tmpDataRec, tmpDataSend, tmp_sim_time, thru2, thru3;
	Packet*  pkptr;
	int in_strm,da1,da2,da3;
	FILE*  fil;
	
	int _block_origin = 0;


	FSM_ENTER (UniApp)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (init) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "init", "UniApp () [init enter execs]")
			{
			/*******************************************************************************************
			** Process Model : Uniapp                                                                 **
			** State         : init                                                                   **
			**                                                                                        **
			** This block intializes the variables that will be used in this process model, loads the **
			** distributions that are used to select the destination address to which the data packet **
			** is sent. It also  gets the object id of the IP module and  the IP address of this host **
			** from that IP module. This also schedules an interrupt for sending data after some init **
			** time. This  init time is given  so that this unicast application doesn't send any data **
			** before all the routing tables are initialized.                                         **
			**                                                                                        **
			*******************************************************************************************/
			
			/*
			** Get the present module's  object ID and load the attributes from  the configuaration file
			*/
			my_objid = op_id_self();
			
			
			op_ima_obj_attr_get( my_objid, "MaxNets", &MAX_NO_NETS );
			op_ima_obj_attr_get( my_objid, "MaxNodes", &MAX_NO_NODES );
			op_ima_obj_attr_get( my_objid, "InitTime", &DataSendInitTime );
			op_ima_obj_attr_get( my_objid,"DATA_IAT", &data_iat);
			
			/*added by Jiemei*/
			/*DataSendInitTime=10;
			data_iat=0.1;*/
			
			/*statistic handler initialization */
			l_pkt_recv_handler = op_stat_reg("L pkt received (pkts)",OPC_STAT_INDEX_NONE, OPC_STAT_LOCAL); 
			l_bit_recv_handler = op_stat_reg("L bit received (bits)",OPC_STAT_INDEX_NONE, OPC_STAT_LOCAL);
			l_delay_handler = op_stat_reg("L end-to-end delay (s)",OPC_STAT_INDEX_NONE, OPC_STAT_LOCAL);
			
			g_pkt_recv_handler = op_stat_reg("G pkt received (pkts)",OPC_STAT_INDEX_NONE, OPC_STAT_GLOBAL);
			g_bit_recv_handler = op_stat_reg("G bit received (bits)",OPC_STAT_INDEX_NONE, OPC_STAT_GLOBAL);
			g_delay_handler = op_stat_reg("G end-to-end delay (s)",OPC_STAT_INDEX_NONE, OPC_STAT_GLOBAL);
			
			g_avg_hops_handler = op_stat_reg("G average hops (hops)",OPC_STAT_INDEX_NONE, OPC_STAT_GLOBAL);
			
			/*statistic variable initialization */
			l_pkt_recv = 0;
			l_bit_recv = 0;
			l_delay = 0;
			
			g_pkt_recv = 0;
			g_bit_recv = 0;
			g_delay = 0;
			
			g_avg_hops = 0;
			
			/*printf("UniApp: MyObjId  %d  MaxNets %d  MaxNodes %d  DataSendInitTime %d  DataIat %f\n",my_objid,MAX_NO_NETS,MAX_NO_NODES,DataSendInitTime,data_iat); */ 
			
			/* 
			** Get the IP module's Object Id  and use this object id to  get the IP address  of the host
			*/
			ip_objid = op_topo_assoc (my_objid, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, 0);
			op_ima_obj_attr_get( ip_objid, "net_num0", &LocalAddress.NetNum );
			op_ima_obj_attr_get( ip_objid, "node_num0", &LocalAddress.NodeNum );
			
			/*
			** Load the distribution pointers for selecting the destination address for the data packet 
			** that is being sent from this state
			*/
			net_dist = op_dist_load("uniform_int", 1, MAX_NO_NETS );
			node_dist = op_dist_load("uniform_int", 0, MAX_NO_NODES );
			
			/*
			** Schedule the interrupt to send the data packets 
			*/
			/*op_intrpt_schedule_self( (op_sim_time() + data_iat + DataSendInitTime), send_data_intr );*/
			}


		/** state (init) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "init", "UniApp () [init exit execs]")
			{
			}


		/** state (init) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (idle) enter executives **/
		FSM_STATE_ENTER_UNFORCED (1, state1_enter_exec, "idle", "UniApp () [idle enter execs]")
			{
			/*added by Jiemei */
			
			/* if (op_sim_time()>10) printf ("UniApp: sent %f receive %f\n",NoOfPktsSent,NoOfPktsRecv); */
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (3,UniApp)


		/** state (idle) exit executives **/
		FSM_STATE_EXIT_UNFORCED (1, state1_exit_exec, "idle", "UniApp () [idle exit execs]")
			{
			}


		/** state (idle) transition processing **/
		FSM_INIT_COND (PK_ARR)
		FSM_TEST_COND (SELF_INT1)
		FSM_TEST_LOGIC ("idle")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 2, state2_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 3, state3_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (pk_arr) enter executives **/
		FSM_STATE_ENTER_FORCED (2, state2_enter_exec, "pk_arr", "UniApp () [pk_arr enter execs]")
			{
			
			/*****************************************************************************************************
			** Process Model : Uniapp                                                                           **
			** State         : pk_arr                                                                           **
			**                                                                                                  **
			** This is a port for all the traffic(packets  with protocol ID as 20) that was destinated for this **
			** host. The state upon receiving the packet updates the counter which keeps track of the number of **
			** packets received  by this host. It writes this statistics into a probe file which can be plotted **
			** once the simulation is completed. It destroys the packets that are received.                     **
			**                                                                                                  **
			*****************************************************************************************************/
			
			/*added by Jiemei, Mar 28, 97*/
			int src_net, src_node, id, treeid;
			App_Data* DataRecv;
			Packet *dataptr;
			char msgtype[10];
			float senttime; 
			char hops;
			double delay;
			
			pkptr  = op_pk_get (op_intrpt_strm ());
			
			/*
			** Probe to measure the number of data packets received by this UDP port
			*/
			PROBE_ON
			{
				l_pkt_recv++;
				g_pkt_recv++;
			
				l_bit_recv += op_pk_total_size_get(pkptr);
				g_bit_recv += op_pk_total_size_get(pkptr);
			
				op_stat_write( l_pkt_recv_handler, l_pkt_recv );
				op_stat_write( g_pkt_recv_handler, g_pkt_recv );
				op_stat_write( l_bit_recv_handler, l_bit_recv );
				op_stat_write( g_bit_recv_handler, g_bit_recv );
			
			/* Added by DG and Jiemei, 11/27/97 */
			
				op_pk_nfd_get(pkptr, "data", &dataptr );
				op_pk_nfd_get(dataptr, "UniData", &DataRecv );
				hops =DataRecv -> Message[0];
				sscanf( &(DataRecv -> Message[1]), "%s %f", msgtype, &senttime );
			printf( "\n=== %d %s %s %f\n", (int)hops, &(DataRecv -> Message[1]), msgtype, senttime );
				delay = op_sim_time() - senttime;
			
				g_avg_hops += hops;
				op_stat_write( g_avg_hops_handler, g_avg_hops / g_pkt_recv );
			
				l_delay += delay;
				g_delay += delay;
				op_stat_write( l_delay_handler, l_delay / l_pkt_recv );
				op_stat_write( g_delay_handler, g_delay / g_pkt_recv );
			
			}
			
			/*added by Jiemei*/
			VL{
			op_pk_nfd_get(pkptr,"src_net",&src_net);
			op_pk_nfd_get(pkptr,"src_node",&src_node);
			printf( "   %fs, (%d,%d) UAPP get packet from (%d, %d)\n", 
				op_sim_time(),LocalAddress.NetNum, LocalAddress.NodeNum, src_net, src_node );
			}
			
			op_prg_mem_free (DataRecv);
			op_pk_destroy (dataptr);
			op_pk_destroy (pkptr);
			
			}


		/** state (pk_arr) exit executives **/
		FSM_STATE_EXIT_FORCED (2, state2_exit_exec, "pk_arr", "UniApp () [pk_arr exit execs]")
			{
			}


		/** state (pk_arr) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (data) enter executives **/
		FSM_STATE_ENTER_FORCED (3, state3_enter_exec, "data", "UniApp () [data enter execs]")
			{
			/*******************************************************************************************
			** Process Model : Uniapp                                                                 **
			** State         : data                                                                   **
			**                                                                                        **
			** This block selects a destination address randomly(from the uniform distributions) from **
			** the networks that are present into the model. It packs the data packet encapsulates it **
			** into the IP datagram and sends the packet to the IP layer. The message is of 500 bytes **
			** length. This block also keeps track of the no. of packets sent by this host and writes **
			** this  statistic into a probe file that can  plotted through the analysis tool. It also **
			** schedules an interrput for sending the next data packet.                               **
			**                                                                                        **
			*******************************************************************************************/
			
			App_Data* DataSent;  /*changed by Jiemei Mar 27, 97*/
			Packet* ip_pkptr;
			int dest_net, dest_node, data_len;
			
			/*
			** Get the destination address from the uniform distributions
			*/
			dest_net = op_dist_outcome( net_dist );
			dest_node = op_dist_outcome( node_dist );
			
			
			DataSent=(App_Data *) op_prg_mem_alloc(sizeof(App_Data));   /*added by Jiemei, Mar 27,97*/
			sprintf( DataSent->Message, "UAPP  %f Src %d %d Dest %d %d\n", op_sim_time(), 
					LocalAddress.NetNum, LocalAddress.NodeNum,
					dest_net, dest_node );	/*added by Jiemei, Mar 27,97*/
			
			
			TRACE VL printf("UApp: %fs (%d.%d)send data to (%d.%d)\n", 
					op_sim_time(), LocalAddress.NetNum, LocalAddress.NodeNum,
					dest_net, dest_node );
			
			/*
			** Create the data packet and load it with data
			*/
			pkptr = op_pk_create_fmt("Unicast_Appl");
			
			op25_pk_nfd_set(pkptr, "UniData", DataSent, op_prg_mem_alloc,op_prg_mem_free, sizeof(App_Data));
			
			data_len = op_pk_total_size_get( pkptr )/8;
			
			/*
			** Encapsulate the data packet into the IP datagram by filling the sender address, destination 
			** address, data length, protocol Id and time-to-live
			*/
			ip_pkptr = op_pk_create_fmt("ip_dgram");
			op25_pk_nfd_set(ip_pkptr, "data", pkptr );
			op25_pk_nfd_set(ip_pkptr, "prot_id", UDP_PROTOCOL_ID );
			op_pk_bulk_size_set(ip_pkptr, data_len*8 );
			op25_pk_nfd_set(ip_pkptr, "dest_net", dest_net );
			op25_pk_nfd_set(ip_pkptr, "dest_node", dest_node );
			op25_pk_nfd_set(ip_pkptr, "src_net", LocalAddress.NetNum );
			op25_pk_nfd_set(ip_pkptr, "src_node", LocalAddress.NodeNum );
			op25_pk_nfd_set(ip_pkptr, "orig_len", data_len );
			op25_pk_nfd_set(ip_pkptr, "frag_len", data_len );
			op25_pk_nfd_set(ip_pkptr, "ttl", 2);   
			op25_pk_nfd_set(ip_pkptr, "frag", 0 );
			
			op_pk_send( ip_pkptr, IP_STREAM );      /*          Send the data packet to the IP module                */
			
			
			/*
			** Write the no. of data packets sent into the probe file 
			*/
			PROBE_ON
			{
			}
			
			/*
			** Schedule the interrupt to send the data packet again
			*/
			op_intrpt_schedule_self( (op_sim_time() + (data_iat)), send_data_intr );
			
			}


		/** state (data) exit executives **/
		FSM_STATE_EXIT_FORCED (3, state3_exit_exec, "data", "UniApp () [data exit execs]")
			{
			}


		/** state (data) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,UniApp)
	}




void
UniApp_svar (prs_ptr,var_name,var_p_ptr)
	UniApp_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (UniApp_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("my_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_my_objid);
	if (Vos_String_Equal ("ip_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_objid);
	if (Vos_String_Equal ("data_iat" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_data_iat);
	if (Vos_String_Equal ("print_time" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_print_time);
	if (Vos_String_Equal ("MAX_NO_NETS" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MAX_NO_NETS);
	if (Vos_String_Equal ("MAX_NO_NODES" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MAX_NO_NODES);
	if (Vos_String_Equal ("DataSendInitTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_DataSendInitTime);
	if (Vos_String_Equal ("net_dist" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_dist);
	if (Vos_String_Equal ("node_dist" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_dist);
	if (Vos_String_Equal ("LocalAddress" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_LocalAddress);
	if (Vos_String_Equal ("l_pkt_recv_handler" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_l_pkt_recv_handler);
	if (Vos_String_Equal ("l_bit_recv_handler" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_l_bit_recv_handler);
	if (Vos_String_Equal ("l_delay_handler" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_l_delay_handler);
	if (Vos_String_Equal ("g_pkt_recv_handler" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_g_pkt_recv_handler);
	if (Vos_String_Equal ("g_bit_recv_handler" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_g_bit_recv_handler);
	if (Vos_String_Equal ("g_delay_handler" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_g_delay_handler);
	if (Vos_String_Equal ("g_avg_hops_handler" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_g_avg_hops_handler);
	if (Vos_String_Equal ("l_pkt_recv" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_l_pkt_recv);
	if (Vos_String_Equal ("l_bit_recv" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_l_bit_recv);
	if (Vos_String_Equal ("l_delay" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_l_delay);

	FOUT;
	}




void
UniApp_diag ()
	{
	double thru, thru1, tmpDataRec, tmpDataSend, tmp_sim_time, thru2, thru3;
	Packet*  pkptr;
	int in_strm,da1,da2,da3;
	FILE*  fil;
	
	int _block_origin;

	FIN (UniApp_diag ())

	BINIT

	FOUT;
	}




void
UniApp_terminate ()
	{
	double thru, thru1, tmpDataRec, tmpDataSend, tmp_sim_time, thru2, thru3;
	Packet*  pkptr;
	int in_strm,da1,da2,da3;
	FILE*  fil;
	
	int _block_origin;

	FIN (UniApp_terminate ())

	BINIT

	FOUT;
	}



Compcode
UniApp_init (pr_state_pptr)
	UniApp_state			**pr_state_pptr;
	{
	int _block_origin = 0;
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	extern int				Vos_Nop ();

	FIN (UniApp_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (UniApp)",
			sizeof (UniApp_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (UniApp_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



enum { _block_origin = __LINE__ };
